"""This module contains dispatched cast() implementations for datetime data."""
# pylint: disable=unused-argument
import datetime
from functools import partial

import numpy as np
import pandas as pd

from pdcast import types
from pdcast.detect import detect_type
from pdcast.resolve import resolve_type
from pdcast.util import time
from pdcast.util.numeric import boundscheck
from pdcast.util.round import round_div, Tolerance
from pdcast.util.vector import apply_with_errors

from .base import cast, generic_to_integer


# TODO: pdcast.cast("1883-11-18 12:00:00", "datetime", tz="US/Pacific")
# induces an `AmbiguousTimeError: Cannot infer dst time from 1883-11-18
# 12:00:00, try using the 'ambiguous' argument`

# -> have to pass in ambiguous_tz, nonexistent_tz arguments to cast()


# TODO: string_to_datetime cannot convert quarterly dates


@cast.overload("datetime", "bool")
def datetime_to_boolean(
    series: pd.Series,
    dtype: types.ScalarType,
    tol: Tolerance,
    rounding: str,
    unit: str,
    step_size: int,
    since: time.Epoch,
    errors: str,
    **unused
) -> pd.Series:
    """Convert timedelta data to a boolean data type."""
    # 2-step conversion: timedelta -> decimal, decimal -> bool
    series = cast(
        series,
        "decimal",
        tol=tol,
        rounding=rounding,
        unit=unit,
        step_size=step_size,
        since=since,
        errors=errors
    )
    return cast(
        series,
        dtype,
        tol=tol,
        rounding=rounding,
        unit=unit,
        step_size=step_size,
        since=since,
        errors=errors,
        **unused
    )


@cast.overload("datetime[pandas]", "int")
def pandas_timestamp_to_integer(
    series: pd.Series,
    dtype: types.ScalarType,
    rounding: str,
    tol: Tolerance,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tz: datetime.tzinfo,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert pandas Timestamps into an integer data type."""
    # apply tz if naive
    if tz and detect_type(series).tz is None:
        series = series.dt.tz_localize(tz)

    # convert to ns (overflow-safe)
    series = series.astype(np.int64)

    # apply epoch
    if since:
        series = series.astype(resolve_type(int).dtype)  # overflow-safe
        series -= since.offset

    # convert ns to final unit, step_size
    if unit != "ns":
        series = time.convert_unit(
            series,
            "ns",
            unit,
            rounding=rounding or "down"
        )
    if step_size != 1:
        series = round_div(
            series,
            step_size,
            rule=rounding or "down"
        )

    # check for overflow
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # delegate to generic conversion
    return generic_to_integer(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors
    )


@cast.overload("datetime[python]", "int")
def python_datetime_to_integer(
    series: pd.Series,
    dtype: types.ScalarType,
    rounding: str,
    tol: Tolerance,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tz: datetime.tzinfo,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert python datetimes into an integer data type."""
    series_type = detect_type(series)

    # apply tz if naive
    if tz and series_type.tz is None:
        series = apply_with_errors(
            series,
            partial(time.localize_pydatetime_scalar, tz=tz),
            errors="raise"
        )

    # convert to ns
    series = apply_with_errors(series, time.pydatetime_to_ns, errors="raise")
    series = series.astype(resolve_type(int).dtype)

    # apply epoch
    if since:
        series -= since.offset

    # convert ns to final unit, step_size
    if unit != "ns":
        series = time.convert_unit(
            series,
            "ns",
            unit,
            rounding=rounding or "down"
        )
    if step_size != 1:
        series = round_div(
            series,
            step_size,
            rule=rounding or "down"
        )

    # check for overflow
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # delegate to generic conversion
    return generic_to_integer(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors
    )


@cast.overload("datetime[numpy]", "int")
def numpy_datetime64_to_integer(
    series: pd.Series,
    dtype: types.ScalarType,
    rounding: str,
    tol: Tolerance,
    unit: str,
    step_size: int,
    since: time.Epoch,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert numpy datetime64s into an integer data type."""
    series_type = detect_type(series)

    # NOTE: using numpy M8 array is ~2x faster than looping through series
    M8_str = f"M8[{series_type.step_size}{series_type.unit}]"
    arr = series.to_numpy(M8_str).view(np.int64).astype(object)

    # correct for M8 step size
    arr *= series_type.step_size

    # convert to final unit
    if since:
        # convert to ns, subtract epoch, and then convert to final unit
        arr = time.convert_unit(
            arr,
            series_type.unit,
            "ns"
        )
        arr -= since.offset
        arr = time.convert_unit(
            arr,
            "ns",
            unit,
            rounding=rounding or "down"
        )

    else:
        # skip straight to final unit
        arr = time.convert_unit(
            arr,
            series_type.unit,
            unit,
            rounding=rounding or "down"
        )

    # apply final step size
    if step_size != 1:
        arr = round_div(arr, step_size, rule=rounding or "down")

    # re-wrap as pandas Series
    series = pd.Series(arr, index=series.index)

    # check for overflow
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # delegate to generic conversion
    return generic_to_integer(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors
    )


@cast.overload("datetime", "float")
def datetime_to_float(
    series: pd.Series,
    dtype: types.ScalarType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tol: Tolerance,
    rounding: str,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert timedelta data to a floating point data type."""
    # 2 step conversion: datetime -> ns, ns -> float
    series = cast(
        series,
        int,
        unit="ns",
        step_size=1,
        since=since,
        rounding=None,
        downcast=None,
        errors=errors
    )

    # convert ns to final unit
    if unit != "ns":
        series = time.convert_unit(
            series,
            "ns",
            unit,
            rounding=rounding,
            since=since
        )
        if rounding is None:
            series = series.astype(resolve_type(float).dtype)

    # apply final step size
    if step_size != 1:
        series /= step_size
        series = series.astype(resolve_type(float).dtype)

    # integer/float -> float
    return cast(
        series,
        dtype,
        unit=unit,
        step_size=step_size,
        since=since,
        tol=tol,
        rounding=rounding,
        downcast=downcast,
        errors=errors,
        **unused
    )


@cast.overload("datetime", "complex")
def datetime_to_complex(
    series: pd.Series,
    dtype: types.ScalarType,
    downcast: types.CompositeType,
    **unused
) -> pd.Series:
    """Convert timedelta data to a complex data type."""
    # 2-step conversion: datetime -> float, float -> complex
    series = cast(series, dtype.equiv_float, downcast=None, **unused)
    return cast(series, dtype, downcast=downcast, **unused)


@cast.overload("datetime", "decimal")
def datetime_to_decimal(
    series: pd.Series,
    dtype: types.ScalarType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tol: Tolerance,
    rounding: str,
    errors: str,
    **unused
) -> pd.Series:
    """Convert timedelta data to a decimal data type."""
    # 2-step conversion: datetime -> ns, ns -> decimal
    series = cast(
        series,
        "int",
        unit="ns",
        step_size=1,
        since=since,
        rounding=None,
        downcast=None,
        errors=errors
    )
    series = cast(
        series,
        dtype,
        unit=unit,
        step_size=step_size,
        since=since,
        tol=tol,
        rounding=rounding,
        errors=errors,
        **unused
    )

    # convert decimal ns to final unit
    if unit != "ns":
        series = time.convert_unit(
            series,
            "ns",
            unit,
            rounding=rounding,
            since=since
        )

    # apply final step size
    if step_size != 1:
        series /= step_size

    return series


@cast.overload("datetime", "datetime")
def datetime_to_datetime(
    series: pd.Series,
    dtype: types.ScalarType,
    rounding: str,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tz: datetime.tzinfo,
    errors: str,
    **unused
) -> pd.Series:
    """Convert datetime data to another datetime representation."""
    series_type = detect_type(series)

    # trivial case
    if dtype == series_type:
        return series

    # 2-step conversion: datetime -> ns, ns -> datetime
    series = cast(
        series,
        "int",
        rounding=rounding,
        unit="ns",
        step_size=1,
        since=time.Epoch("utc"),
        tz=tz,
        downcast=None,
        errors=errors
    )
    return cast(
        series,
        dtype,
        rounding=rounding,
        unit="ns",
        step_size=1,
        since=time.Epoch("utc"),
        tz=tz,
        errors=errors,
        **unused
    )


@cast.overload("datetime[pandas]", "datetime[pandas]")
def pandas_timestamp_to_pandas_timestamp(
    series: pd.Series,
    dtype: types.ScalarType,
    tz: datetime.tzinfo,
    **unused
) -> pd.Series:
    """Fastpath for same-class pandas timestamp conversions."""
    series_type = detect_type(series)

    # reconcile time zones
    if tz:
        dtype = dtype.replace(tz=tz)

    # trivial case
    if series_type == dtype:
        return series

    # localize/convert tim ezones
    if not series_type.tz:
        return series.dt.tz_localize(dtype.tz)
    return series.dt.tz_convert(dtype.tz)


@cast.overload("datetime[python]", "datetime[python]")
def python_datetime_to_python_datetime(
    series: pd.Series,
    dtype: types.ScalarType,
    tz: datetime.tzinfo,
    **unused
) -> pd.Series:
    """Fastpath for same-class pandas timestamp conversions."""
    # reconcile time zones
    if tz:
        dtype = dtype.replace(tz=tz)

    # trivial case: time zones are identical
    if detect_type(series) == dtype:
        return series

    # localize/convert time zones
    call = partial(time.localize_pydatetime_scalar, tz=tz)
    result = series.apply(call, convert_dtype=False)
    return result.astype(dtype.dtype)


@cast.overload("datetime", "timedelta")
def datetime_to_timedelta(
    series: pd.Series,
    dtype: types.ScalarType,
    unit: str,
    step_size: int,
    rounding: str,
    since: time.Epoch,
    errors: str,
    **unused
) -> pd.Series:
    """Convert datetime data to a timedelta representation."""
    # 2-step conversion: datetime -> ns, ns -> timedelta
    series = cast(
        series,
        "int",
        unit="ns",
        step_size=1,
        rounding=rounding,
        since=since,
        downcast=None,
        errors=errors
    )
    return cast(
        series,
        dtype,
        unit="ns",
        step_size=1,
        rounding=rounding,
        since=since,
        errors=errors,
        **unused
    )



import pandas as pd

from pdcast.check import typecheck
from pdcast.convert import cast
from pdcast.detect import detect_type

from .round import round
from .dt.tz_convert import tz_convert
from .dt.tz_localize import tz_localize


# ignore this file when doing frame-based object lookups in resolve_type()
IGNORE_FRAME_OBJECTS = True


def attach() -> None:
    """Grouped :meth:`attach_to <pdcast.Attachable.attach_to>` operations.
    """
    # Series
    cast.attach_to(pd.Series)
    detect_type.attach_to(pd.Series, pattern="property", name="element_type")
    typecheck.attach_to(pd.Series)
    round.attach_to(pd.Series)
    tz_localize.attach_to(pd.Series, namespace="dt")
    tz_convert.attach_to(pd.Series, namespace="dt")

    # DataFrame
    cast.attach_to(pd.DataFrame)
    detect_type.attach_to(pd.DataFrame, pattern="property", name="element_type")
    typecheck.attach_to(pd.DataFrame)


def detach() -> None:
    """Grouped :meth:`detach <pdcast.Attachable.detach>` operations.
    """
    # pylint: disable=no-member

    # Series
    pd.Series.cast.detach()
    pd.Series.element_type.detach()
    pd.Series.typecheck.detach()
    pd.Series.round.detach()
    pd.Series.dt.tz_localize.detach()
    pd.Series.dt.tz_convert.detach()

    # DataFrame
    pd.DataFrame.cast.detach()
    pd.DataFrame.element_type.detach()
    pd.DataFrame.typecheck.detach()



"""This module contains utility functions for manipulating datetime-like data
in a variety of formats.

Functions
---------
ns_to_pydatetime()
    Convert an integer vector of nanoseconds from the utc epoch into
    corresponding ``datetime.datetime`` objects.

pydatetime_to_ns()
    Convert python ``datetime.datetime`` objects into an integer number of
    nanoseconds from the utc epoch.

numpy_datetime64_to_ns()
    Convert ``numpy.datetime64`` objects into an integer number of nanoseconds
    from the utc epoch.

is_iso_8601_format_string()
    Check whether an ``strftime()``-compliant format string is ISO
    8601-compliant.

iso_8601_to_ns()
    Convert a vector of ISO 8601 datetime strings into an integer number of
    nanoseconds from the utc epoch.

string_to_pydatetime()
    Convert datetime strings into python ``datetime.datetime`` objects using
    ``dateutil``.

filter_dateutil_parser_error()
    Distinguish between malformed values and overflow in ``dateutil`` parser
    errors.
"""
from cpython cimport datetime
import datetime
import re

import dateutil
import numpy as np
cimport numpy as np
import pandas as pd
import pytz
import tzlocal

from .calendar import date_to_days, days_in_month
from .timedelta cimport pytimedelta_to_ns
# from .timezone import localize_pydatetime_scalar
from .unit cimport as_ns


################################
####    PANDAS TIMESTAMP    ####
################################


cpdef inline object pandas_timestamp_to_ns(object date, object tz = None):
    """Convert a pandas Timestamp into a nanosecond offset from UTC."""
    if tz and not date.tzinfo:
        date = date.tz_localize(tz)

    # NOTE: as of pandas 2.0, Timestamps can hold non-ns units

    if date.unit == "ns":
        return date.value
    return int(date.asm8.view("i8")) * as_ns[date.unit]


###############################
####    PYTHON DATETIME    ####
###############################


cdef datetime.datetime py_naive_utc = datetime.datetime.utcfromtimestamp(0)
cdef datetime.datetime py_aware_utc = (
    py_naive_utc.replace(tzinfo=datetime.timezone.utc)
)


cpdef inline datetime.datetime ns_to_pydatetime(
    object ns,
    object tz = None
):
    """Convert a nanosecond offset from UTC into a properly-localized
    `datetime.datetime` object.
    """
    cdef datetime.timedelta offset

    offset = datetime.timedelta(microseconds=int(ns // as_ns["us"]))
    if tz is None:
        return py_naive_utc + offset

    cdef datetime.datetime result = py_aware_utc + offset

    return result if is_utc(tz) else result.astimezone(tz)


cpdef inline object pydatetime_to_ns(datetime.datetime date, object tz = None):
    """Convert a python datetime into a nanosecond offset from UTC."""
    if tz and not date.tzinfo:
        if isinstance(tz, pytz.BaseTzInfo):  # use .localize
            date = tz.localize(date)
        else:  # use .replace
            date = date.replace(tzinfo=tz)

    cdef datetime.timedelta result

    if date.tzinfo:
        result = date - py_aware_utc
    else:
        result = date - py_naive_utc

    return pytimedelta_to_ns(result)


################################
####    NUMPY DATETIME64    ####
################################


cpdef object numpy_datetime64_to_ns(
    object date,
    str unit = None,
    long int step_size = -1
):
    """Convert a numpy datetime64 into a nanosecond offset from UTC."""
    if unit is None or step_size < 0:
        unit, step_size = np.datetime_data(date)

    cdef object result

    result = int(date.view(np.int64)) * step_size
    if unit == "ns":
        return result
    if unit in as_ns:
        return result * as_ns[unit]
    if unit == "M":
        return date_to_days(1970, 1 + result, 1) * as_ns["D"]
    return date_to_days(1970 + result, 1, 1) * as_ns["D"]


########################
####    TIMEZONE    ####
########################


cdef datetime.tzinfo utc = datetime.timezone.utc


cpdef datetime.datetime localize_pydatetime_scalar(
    datetime.datetime dt,
    object tz
):
    """Localize a scalar datetime.datetime object to the given tz."""
    # datetime is naive
    if not dt.tzinfo:
        if tz is None:
            return dt  # do nothing

        # localize directly to final tz
        if isinstance(tz, pytz.BaseTzInfo):  # use .localize
            return tz.localize(dt)
        return dt.replace(tzinfo=tz)

    # datetime is aware
    if tz is None:  # convert to utc, then strip tzinfo
        return dt.astimezone(utc).replace(tzinfo=None)
    return dt.astimezone(tz)  # convert to final tz


cpdef inline bint is_utc(datetime.tzinfo tz):
    """Check whether a tzinfo object corresponds to UTC."""
    return tz.utcoffset(None) == datetime.timedelta(0)


######################
####    STRING    ####
######################


cdef object build_iso_8601_regex():
    """Compile a regex pattern to match arbitrary ISO-8601 datetime strings."""
    # extract date component
    cdef str sign = r"(?P<sign>-)"
    cdef str year = r"(?P<year>[0-9]+)"
    cdef str month = r"(?P<month>[0-9]{2})"
    cdef str day = r"(?P<day>[0-9]{2})"

    # ISO 8601 date/time separators
    cdef str separators = r"[T\s]"

    # extract time component
    cdef str hour = r"(?P<hour>[0-9]{2})"
    cdef str minute = r"(?P<minute>[0-9]{2})"
    cdef str second = r"(?P<second>[0-9]{2}\.?[0-9]*)"

    # extract utc offset
    cdef str utc_sign = r"(?P<utc_sign>[+-])"
    cdef str utc_hour = r"(?P<utc_hour>[0-9]{2})"
    cdef str utc_minute = r"(?P<utc_minute>[0-9]{2})"

    # combine
    return re.compile(
        rf"^{sign}?{year}-?{month}?-?{day}?"            # date
        rf"{separators}?"                               # separator
        rf"{hour}?:?{minute}?:?{second}?"               # time
        rf"(Z|{utc_sign}{utc_hour}:?{utc_minute})?$"    # utc offset
    )


cdef object build_iso_8601_strptime_format_regex():
    """Compile a regex to match strptime/strftime format strings if they
    are ISO 8601-compliant.
    """
    opt = lambda x: rf"({x})?"

    # work from right to left to fully account for optional components
    cdef str result = r"(%z|Z)?"
    result = rf"(.%f{result})?"
    result = rf"(:%S{result})?"
    result = rf"(:%M{result})?"
    result = rf"(( |T)%H{result})?"
    result = rf"(-%d{result})?"
    result = rf"(-%m{result})?"
    result = rf"^%Y{result}$"
    # reference: "%Y-%m-%d %H:%M:%S.%f%z"

    return re.compile(result)


cdef object iso_8601_format_pattern = build_iso_8601_strptime_format_regex()
cdef object iso_8601_pattern = build_iso_8601_regex()


cpdef inline bint is_iso_8601_format_string(str input_string):
    """Returns `True` if an strftime/strptime format string complies with the
    ISO 8601 standard.
    """
    return iso_8601_format_pattern.match(input_string) is not None


cpdef object iso_8601_to_ns(str input_string):
    """Convert a scalar ISO 8601 string into a nanosecond offset from the
    utc epoch ('1970-01-01 00:00:00+0000').

    Returns a 2-tuple with the nanosecond offset as the first index.  The
    second index contains a boolean indicating whether the string had a
    timezone specifier (either 'Z' or a valid UTC offset).
    """
    cdef object match = iso_8601_pattern.match(input_string)
    if not match:
        raise ValueError(f"Invalid isoformat string: {repr(input_string)}")

    cdef dict parts = match.groupdict()

    # extract date components
    cdef char sign = -1 if parts["sign"] else 1
    cdef long int year = int(parts["year"])
    cdef short month = int(parts["month"] or 1)
    cdef short day = int(parts["day"] or 1)

    # extract time components
    cdef long int hour = int(parts["hour"] or 0)
    cdef long int minute = int(parts["minute"] or 0)
    cdef double second = float(parts["second"] or 0)

    # extract utc offset components
    cdef char utc_sign = -1 if parts["utc_sign"] == "-" else 1
    cdef long int utc_hour = int(parts["utc_hour"] or 0)
    cdef long int utc_minute = int(parts["utc_minute"] or 0)

    # check values are within normal bounds
    if not (
        1 <= month <= 12 and
        1 <= day <= days_in_month(month, year) and
        0 <= hour < 24 and
        0 <= minute < 60 and
        0 <= second < 60 and
        0 <= utc_hour < 24 and
        0 <= utc_minute < 60
    ):
        raise ValueError(f"invalid isoformat string {repr(input_string)}")

    cdef object result

    # convert date to ns, add time component, and subtract utc offset
    result = date_to_days(sign * year, month, day) * as_ns["D"]
    result += hour * as_ns["h"] + minute * as_ns["m"] + int(second * as_ns["s"])
    result -= utc_sign * (utc_hour * as_ns["h"] + utc_minute * as_ns["m"])
    return result


cpdef datetime.datetime string_to_pydatetime(
    str input_string,
    str format = None,
    object parser_info = None,
    object tz = None,
    str errors = "raise"
):
    """Convert a string to a python datetime object."""
    input_string = input_string.strip().lower()
    if input_string.startswith("-"):
        raise OverflowError(
            f"datetime.datetime objects cannot represent negative years"
        )

    cdef datetime.datetime result = None

    # attempt to use format string if given
    if format is not None:
        try:
            result = datetime.datetime.strptime(input_string, format)
        except ValueError as err:
            if errors != "coerce":
                raise err            

    # check for relative date
    if result is None and input_string in ("today", "now"):
        result = datetime.datetime.now()

    # check for quarterly date
    if result is None and "q" in input_string:
        try:
            period = pd.Period(input_string, freq="Q") - 1
            days = date_to_days(period.year, 1, 1 + period.day_of_year)
            result = ns_to_pydatetime(days * as_ns["D"])
        except pd._libs.tslibs.parsing.DateParseError:
            pass

    # parse using dateutil
    if result is None:
        try:
            result = dateutil.parser.parse(
                input_string,
                fuzzy=True,
                parserinfo=parser_info
            )
        except dateutil.parser.ParserError as err:
            raise filter_dateutil_parser_error(err) from None

    # apply timezone
    return localize_pydatetime_scalar(result, tz=tz)


####################
####    MISC    ####
####################


cdef object parser_overflow_pattern = re.compile(r"out of range|must be in")


cpdef Exception filter_dateutil_parser_error(Exception err):
    """Convenience function to differentiate dateutil overflow errors from
    those that are raised due to malformed values.
    """
    cdef str err_msg = str(err)
    if parser_overflow_pattern.search(err_msg):
        return OverflowError(err_msg)
    return ValueError(err_msg)



"""This module contains all the prepackaged float types for the ``pdcast``
type system.
"""
import numpy as np

from .base import Type, TypeMeta


class Float(Type):
    """Abstract float type."""

    aliases = {"float", "floating", "f"}

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import Complex
        return Complex


@Float.default
class Float64(Float):
    """Abstract 64-bit float type."""

    aliases = {"float64", "double", "float_", "f8", "d"}
    max = 2**53
    min = -2**53

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import Complex128
        return Complex128


@Float64.default
class NumpyFloat64(Float64, backend="numpy"):
    """Numpy 64-bit float type."""

    aliases = {np.float64, np.dtype(np.float64)}
    dtype = np.dtype(np.float64)
    missing = np.nan

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import NumpyComplex128
        return NumpyComplex128


class PythonFloat(Float, backend="python"):
    """Python float type."""

    aliases = {float}
    scalar = float
    missing = np.nan

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import PythonComplex
        return PythonComplex


class Float32(Float):
    """Abstract 32-bit float type."""

    aliases = {"float32", "single", "f4"}
    max = 2**24
    min = -2**24

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import Complex64
        return Complex64


@Float32.default
class NumpyFloat32(Float32, backend="numpy"):
    """Numpy 32-bit float type."""

    aliases = {np.float32, np.dtype(np.float32)}
    dtype = np.dtype(np.float32)
    missing = np.nan

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import NumpyComplex64
        return NumpyComplex64


class Float16(Float):
    """Abstract 16-bit float type."""

    aliases = {"float16", "half", "f2", "e"}
    max = 2**11
    min = -2**11

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import Complex64
        return Complex64


@Float16.default
class NumpyFloat16(Float16, backend="numpy"):
    """Numpy 16-bit float type."""

    aliases = {np.float16, np.dtype(np.float16)}
    dtype = np.dtype(np.float16)
    missing = np.nan

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import NumpyComplex64
        return NumpyComplex64


#################################
####    LONG DOUBLE (x86)    ####
#################################


# TODO: figure out conditional types?


# NOTE: long doubles are platform-specific and may not be valid depending on
# hardware configuration.


has_longdouble: bool = np.dtype(np.longdouble).itemsize > 8


class Float80(Float):
    """Abstract 80-bit float type."""

    aliases = {
        "float80", "longdouble", "longfloat", "long double", "long float", "f10", "g"
    }

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import Complex160
        return Complex160


@Float80.default
class NumpyFloat80(Float80, backend="numpy"):
    """Numpy 80-bit float type."""

    aliases = {np.longdouble, np.dtype(np.longdouble)}
    dtype = np.dtype(np.longdouble)
    missing = np.nan
    max = 2**64
    min = -2**64

    @property
    def equiv_complex(self) -> TypeMeta:
        """An equivalent complex type."""
        from .complex import NumpyComplex160
        return NumpyComplex160



"""This module describes a ``Tolerance`` object, which stores a positive
complex value as two separate ``Decimal`` values.
"""
import decimal

import numpy as np

from pdcast.util.type_hints import numeric


cdef class Tolerance:
    """A simple struct describing a complex tolerance for precision loss
    operations.

    Parameters
    ----------
    tol : numeric
        A positive numeric to convert into a
        :class:`Decimal <python:decimal.Decimal>`\-based :class:`Tolerance`
        object.  If this is a complex value, then its real and imaginary
        components will be considered separately.

    Returns
    -------
    Tolerance
        A complex value that stores its real and imaginary components as
        :class:`Decimal <python:decimal.Decimal>` objects.

    Raises
    ------
    TypeError
        if ``tol`` could not be coerced to a scalar
        :class:`Decimal <python:decimal.Decimal>`.
    ValueError
        If the real or imaginary component of ``tol`` is not positive.
    """

    def __init__(self, tol: numeric):
        # extract real, imaginary components
        real = np.real(tol)
        if isinstance(tol, (complex, np.complexfloating)):
            imag = np.imag(tol)
        else:
            imag = real

        # coerce to decimal
        try:
            self.real = decimal.Decimal(real)
            self.imag = decimal.Decimal(imag)
        except TypeError:
            raise TypeError(f"value cannot be coerced to decimal: {tol}")

        # check positive
        if self.real < 0 or self.imag < 0:
            raise ValueError(f"`tol` must be a numeric >= 0, not {tol}")

    def __bool__(self) -> bool:
        return bool(self.real) or bool(self.imag)

    def __int__(self) -> int:
        return int(self.real)

    def __float__(self) -> float:
        return float(self.real)

    def __complex__(self) -> complex:
        return complex(self.real, self.imag)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.real}+{self.imag}j)"

    def __str__(self) -> str:
        return f"{self.real}+{self.imag}j"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Tolerance):
            return NotImplemented

        return self.real == other.real and self.imag == other.imag




"""This module contains utility functions to help format errors raised by
`pdcast` internals.
"""
from .type_hints import list_like


def shorten_list(seq: list_like, max_length: int = 5) -> str:
    """Converts a list-like into an abridged string for use in error messages.
    """
    if len(seq) <= max_length:
        return str(list(seq))
    shortened = ", ".join(str(i) for i in seq[:max_length])
    return f"[{shortened}, ...] ({len(seq)})"



"""This module contains helper functions for numeric conversions."""
import pandas as pd
cimport numpy as np
import numpy as np

from pdcast cimport types
from pdcast.detect import detect_type
from pdcast.util.error import shorten_list
from pdcast.util.round cimport Tolerance


# TODO: downcast_integer should use its tol argument for bounds checks


# TODO: downcast methods should use the target dtype rather than the series
# type to get smaller types.  Right now, the observed series type will always
# be empty.


######################
####    PUBLIC    ####
######################


cpdef object boundscheck(
    object series,
    types.ScalarType dtype,
    object tol,
    str errors
):
    """Ensure that a series fits within the allowable range of a given data
    type.

    If overflow is detected, this function will attempt to upcast the data
    type to fit the series.  If this fails and errors="coerce", then it
    will drop overflowing values from the series to fit the data type instead.

    Parameters
    ----------
    dtype : ScalarType
        An ScalarType whose range will be used for the check.
    errors : str, default "raise"
        The error-handling rule to apply to the range check.  Must be one
        of "raise", "ignore", or "coerce".

    Returns
    -------
    series : pd.Series
        A series whose elements fit within the range of the specified type.
        In most cases, this will be the original series, but if overflow is
        detected and errors="coerce", then it may be a subset of the original.

    Raises
    ------
    OverflowError
        If ``dtype`` cannot fit the observed range of the series, cannot
        be upcasted to fit, and ``errors != "coerce"``
    """
    cdef types.VectorType series_type = detect_type(series)

    # trivial case for empty series
    if series_type is None:
        return series

    # get min/max values as python ints (prevents inconsistent comparison)
    cdef object min_val = series.min()
    cdef object max_val = series.max()
    cdef object index

    min_val = int(min_val) - bool(min_val % 1)
    max_val = int(max_val) + bool(max_val % 1)

    # check for overflow
    if min_val < dtype.min or max_val > dtype.max:
        # clip to within tolerance
        if min_val >= dtype.min - tol and max_val <= dtype.max + tol:
            return series.clip(dtype.min, dtype.max)

        # continue with OverflowError
        index = (series < dtype.min) | (series > dtype.max)
        if errors == "coerce":
            series = series[~index]
        else:
            raise OverflowError(
                f"values exceed {dtype} range at index "
                f"{shorten_list(series[index].index.values)}"
            )

    return series


cpdef object downcast_integer(
    object series,
    Tolerance tol,
    types.CompositeType smallest
):
    """Reduce the itemsize of an integer type to fit the observed range."""
    from pdcast import convert

    cdef types.VectorType series_type = detect_type(series)
    cdef list smaller = list(series_type.smaller)
    cdef object min_val
    cdef object max_val

    # filter based on `smallest`
    if smallest is not None:
        smaller = filter_smallest(smaller, smallest, series_type=series_type)

    min_val = series.min()
    max_val = series.max()

    # get observed range as python ints (prevents inconsistent comparison)
    if pd.isna(min_val):  # series is empty
        # NOTE: we swap min/max to maintain upcast behavior for generic types
        min_val = series_type.max
        max_val = series_type.min
    else:
        min_val = int(min_val)
        max_val = int(max_val)

    # search for smaller data type that fits observed range
    for small in smaller:
        # range check
        if min_val < small.min or max_val > small.max:
            continue

        # dtype is valid
        return convert.cast(
            series,
            dtype=small,
            downcast=None,
            errors="raise"
        )

    # series could not be downcasted
    return series


cpdef object downcast_float(
    object series,
    Tolerance tol,
    types.CompositeType smallest
):
    """Reduce the itemsize of a float type to fit the observed range."""
    from pdcast import convert

    cdef types.VectorType series_type = detect_type(series)
    cdef list smaller = list(series_type.smaller)

    # filter based on `smallest`
    if smallest is not None:
        smaller = filter_smallest(smaller, smallest, series_type=series_type)

    # try converting to each candidate in order (applying tolerance)
    for small in smaller:
        try:
            attempt = convert.cast(
                series,
                dtype=small,
                tol=tol,
                downcast=None,
                errors="raise"
            )
        except Exception:
            continue

        # candidate is valid
        if within_tol(attempt, series, tol=tol.real).all():
            return attempt

    # return original
    return series


cpdef object downcast_complex(
    object series,
    Tolerance tol,
    types.CompositeType smallest
):
    """Reduce the itemsize of a complex type to fit the observed range."""
    cdef object real_part
    cdef object imag_part

    # downcast real and imaginary component separately
    real_part = downcast_float(
        real(series),
        tol=tol,
        smallest=smallest
    )
    imag_part = downcast_float(
        imag(series),
        tol=Tolerance(tol.imag),
        smallest=smallest
    )

    # use whichever type is larger
    return combine_real_imag(real_part, imag_part)


cpdef np.ndarray[np.uint8_t, cast=True] isinf(object series):
    """Return a boolean mask indicating the position of infinities in the
    series.

    Parameters
    ----------
    series : pd.Series
        A pandas series to check for infinities.

    Returns
    -------
    np.ndarray
        A numpy boolean array indicating the index of each infinity within the
        series.
    """
    return np.isin(series, (np.inf, -np.inf))


cpdef object real(object series):
    """Get the real component of a series.

    This is a convenience function that mimics the behavior of `np.real()`, but
    supports complex values stored in object arrays.

    Parameters
    ----------
    series : pd.Series
        A pandas series to decompose.

    Returns
    -------
    pd.Series
        The real component of the series.
    """
    cdef types.VectorType series_type
    cdef types.VectorType target

    # NOTE: np.real() fails when applied to object arrays that may contain
    # complex values.  In this case, we reduce it to a loop.

    # object array
    if pd.api.types.is_object_dtype(series):
        series_type = detect_type(series)
        target = getattr(series_type, "equiv_float", series_type)
        return elementwise_real(series).astype(target.dtype)

    # use np.real() directly
    return pd.Series(np.real(series), index=series.index)


cpdef object imag(object series):
    """Get the imaginary component of the wrapped series.

    This is a convenience function that mimics the behavior of `np.imag()`, but
    support complex values stored in object arrays.

    Parameters
    ----------
    series : pd.Series
        A pandas series to decompose.

    Returns
    -------
    pd.Series
        The imaginary component of the series.
    """
    cdef types.VectorType series_type
    cdef types.VectorType target

    # NOTE: np.imag() fails when applied to object arrays that may contain
    # complex values.  In this case, we reduce it to a loop.

    # object array
    if pd.api.types.is_object_dtype(series):
        series_type = detect_type(series)
        target = getattr(series_type, "equiv_float", series_type)
        return elementwise_real(series).astype(target.dtype)

    # use np.imag() directly
    return pd.Series(np.imag(series), index=series.index)


cpdef np.ndarray[np.uint8_t, cast=True] within_tol(
    object series1,
    object series2,
    object tol
):
    """Return a boolean mask indicating where the elements of two series are
    within the specified tolerance of one another.

    Parameters
    ----------
    series1 : pd.Series
        A series to compare against ``series2``.
    series2 : pd.Series
        A series to compare against ``series1``.
    tol : numeric
        The maximum tolerance between values.  If any elements of the provided
        series objects differ by more than this amount, then the corresponding
        result will be set ``False``.

    Returns
    -------
    np.ndarray
        A boolean mask indicating which elements of ``series1`` are within
        tolerance of ``series2``.
    """
    cdef object result

    # fastpath if no tolerance is given
    if not tol:
        result = (series1 == series2)
    else:
        result = ~((series1 - series2).abs() > tol)

    return result.to_numpy()


#######################
####    PRIVATE    ####
#######################


cdef object elementwise_real = np.frompyfunc(np.real, 1, 1)
cdef object elementwise_imag = np.frompyfunc(np.imag, 1, 1)


cdef list filter_smallest(
    list smaller,
    types.CompositeType smallest,
    types.VectorType series_type
):
    """Filter a list of downcast candidates based on a set of lower limits."""
    # trivial case: series type is directly contained in lower limits
    if series_type in smallest:
        return []

    cdef list filtered = []

    # iterate from largest to smallest
    for typ in reversed(smaller):
        filtered.append(typ)

        # if candidate is in lower limits, break
        if typ in smallest:
            break

    return list(reversed(filtered))


cdef object combine_real_imag(
    object real_part,
    object imag_part
):
    cdef types.VectorType real_type = detect_type(real_part)
    cdef types.VectorType imag_type = detect_type(imag_part)
    cdef types.VectorType largest
    cdef types.VectorType target
    cdef object result

    largest = max([real_type, imag_type])
    target = largest.equiv_complex
    result = (real_part + imag_part * 1j)
    return result.astype(target.dtype, copy=False)




"""Run Bertrand from the command line to get include directory, version number, etc.
"""
import argparse


from . import __version__
from .setuptools import quick_include


# TODO: Add linker options to `python -m bertrand -I` for pcre2, cpptrace, and googletest


def main() -> None:
    """Run Bertrand as a command-line utility."""
    parser = argparse.ArgumentParser(description="Bertrand module utilities.")
    parser.add_argument(
        "-v", "--version",
        action="store_true",
        help="Print the installed version number."
    )
    parser.add_argument(
        "-I", "--include",
        action="store_true",
        help=(
            "List all the include and link symbols needed to compile a pure-C++ "
            "project that relies on Bertrand as a dependency.  This includes the "
            "Python development headers, numpy, pybind11, pcre2, googletest, and "
            "cpptrace as well as those of bertrand itself.  Users can quickly include "
            "all of these in a single command by adding `$(python3 -m bertrand -I)` "
            "to their compilation flags."
        )
    )

    args = parser.parse_args()
    show_help = True

    if args.include:
        print(" ".join(quick_include()))
        show_help = False

    if args.version:
        print(__version__)
        show_help = False

    if show_help:
        parser.print_help()


if __name__ == "__main__":
    main()



"""This module contains all the prepackaged boolean types for the ``pdcast``
type system.
"""
import numpy as np
import pandas as pd

from .base import Type


class Bool(Type):
    """Abstract boolean type.

    *   **aliases:** ``"bool"``, ``"boolean"``, ``"bool_"``, ``"bool8"``,
        ``"b1"``, ``"?"``
    *   **arguments:** [backend]

        *   **backends:** :class:`numpy <NumpyBooleanType>`,
            :class:`pandas <PandasBooleanType>`,
            :class:`python <PythonBooleanType>`

    *   **type_def:** :class:`bool <python:bool>`
    *   **dtype:** ``numpy.dtype(bool)``
    *   **itemsize:** 1
    *   **na_value**: :class:`pandas.NA`
    *   **is_nullable:** False
    *   **max**: 1
    *   **min**: 0

    .. doctest::

        >>> resolve_type("bool")
        BooleanType()
        >>> resolve_type("boolean[numpy]")
        NumpyBooleanType()
        >>> resolve_type("b1[pandas]")
        PandasBooleanType()
        >>> resolve_type("?[python]")
        PythonBooleanType()
    """

    aliases = {"bool", "boolean", "bool_", "bool8", "b1", "?"}


@Bool.default
class NumpyBool(Bool, backend="numpy"):
    """Numpy boolean type.

    *   **aliases:** :class:`numpy.bool_`, ``numpy.dtype(bool)``
    *   **arguments:** []
    *   **type_def:** :class:`numpy.bool_`
    *   **dtype:** ``numpy.dtype(bool)``
    *   **itemsize:** 1
    *   **na_value**: :class:`pandas.NA`
    *   **is_nullable:** False
    *   **max**: 1
    *   **min**: 0

    .. testsetup::

        import numpy as np
        import pdcast

    .. doctest::

        >>> resolve_type(np.bool_)
        NumpyBooleanType()
        >>> resolve_type(np.dtype(bool))
        NumpyBooleanType()
    """

    aliases = {np.bool_, np.dtype(np.bool_)}
    dtype = np.dtype(np.bool_)
    itemsize = 1
    max = 1
    min = 0
    is_nullable = False


@NumpyBool.nullable
class PandasBool(Bool, backend="pandas"):
    """Pandas boolean type.

    *   **aliases:** ``"Boolean"``, :class:`pandas.BooleanDtype`
    *   **arguments:** []
    *   **type_def:** :class:`numpy.bool_`
    *   **dtype:** :class:`pandas.BooleanDtype() <pandas.BooleanDtype>`
    *   **itemsize:** 1
    *   **na_value**: :class:`pandas.NA`
    *   **is_nullable:** True
    *   **max**: 1
    *   **min**: 0

    .. testsetup::

        import pandas as pd
        import pdcast

    .. doctest::

        >>> resolve_type("Boolean")
        PandasBooleanType()
        >>> resolve_type(pd.BooleanDtype)
        PandasBooleanType()
        >>> resolve_type(pd.BooleanDtype())
        PandasBooleanType()
    """

    aliases = {pd.BooleanDtype}
    dtype = pd.BooleanDtype()
    itemsize = 1
    max = 1
    min = 0


class PythonBool(Bool, backend="python"):
    """Python boolean type.

    *   **aliases:** :class:`bool <python:bool>`
    *   **arguments:** []
    *   **type_def:** :class:`bool <python:bool>`
    *   **dtype:** auto-generated
    *   **itemsize:** 28
    *   **na_value**: :class:`pandas.NA`
    *   **is_nullable:** True
    *   **max**: 1
    *   **min**: 0

    .. testsetup::

        import pdcast

    .. doctest::

        >>> resolve_type(bool)
        PythonBooleanType()
    """

    aliases = {bool}
    scalar = bool
    max = 1
    min = 0



"""This module contains the base class for all bertrand decorator types.

The entire contents of this file are hosted in the online documentation, and serve as
a reference for defining and registering new decorator types with the bertrand type
system.  Users can refer to this file to learn how to build their own decorator types,
and to see how the built-in decorator types are implemented.
"""
from __future__ import annotations

from collections import deque
from typing import Any

import pandas as pd

from .meta import Base, DecoratorMeta


class DecoratorType(Base, metaclass=DecoratorMeta):
    """TODO
    """

    # pylint: disable=missing-param-doc, missing-return-doc, missing-raises-doc

    # TODO: __class_getitem__(), from_scalar(), from_dtype(), from_string()

    @classmethod
    def transform(cls, series: pd.Series[Any]) -> pd.Series[Any]:
        """Base implementation of the transform() method, which is inherited by all
        descendant types.
        """
        return series.astype(cls.dtype, copy=False)

    @classmethod
    def inverse_transform(cls, series: pd.Series[Any]) -> pd.Series[Any]:
        """Base implementation of the inverse_transform() method, which is inherited by
        all descendant types.
        """
        return series.astype(cls.wrapped.dtype, copy=False)

    @classmethod
    def replace(cls, *args: Any, **kwargs: Any) -> DecoratorMeta:
        """Copied from TypeMeta."""
        forwarded = dict(cls.params)
        positional = deque(args)
        n  = len(args)
        i = 0
        for k in forwarded:
            if i < n:
                i += 1
                forwarded[k] = positional.popleft()
            elif k in kwargs:
                forwarded[k] = kwargs.pop(k)

        if positional:
            raise TypeError(
                f"replace() takes at most {len(cls.params)} positional arguments but "
                f"{n} were given"
            )
        if kwargs:
            singular = len(kwargs) == 1
            raise TypeError(
                f"replace() got {'an ' if singular else ''}unexpected keyword argument"
                f"{'' if singular else 's'} [{', '.join(repr(k) for k in kwargs)}]"
            )

        return cls.base_type[*forwarded.values()]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pdcast'
copyright = '2023, Eric Erkela'
author = 'Eric Erkela'
release = '0.0.1'


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_rtd_theme",
    "sphinx_tabs.tabs",
    "sphinx_togglebutton",
]

templates_path = ['_templates']
exclude_patterns = []

intersphinx_mapping = {
    "cython": ("https://cython.readthedocs.io/en/latest/", None),
    "dateutil": ("https://dateutil.readthedocs.io/en/stable/", None),
    "numba": ("https://numba.pydata.org/numba-doc/latest/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "python": ("https://docs.python.org/3", None),
}

# executed before every .. doctest:: block
doctest_global_setup = """
from pdcast import *
"""

# compatible builders for sphinx-tabs
sphinx_tabs_valid_builders = [
    "linkcheck",
]

# config for sphinx-togglebutton
togglebutton_hint = "Click to show"  # hint text for collapsible sections
togglebutton_hint_hide = "Click to hide"  # hint text for collapsed sections
togglebutton_open_on_print = True  # open all collapsible sections during print


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    "navigation_depth": 4,
}
html_static_path = ['_static']
html_css_files = [
    "autosummary_wrap.css",  # force autosummary tables to wrap descriptions
    "collapsible_arrows.css",  # change color of sphinx-togglebutton arrows
]


# -- Options for EPUB output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#epub-options

epub_show_urls = 'footnote'



"""This module describes a pair of functions that can be used to convert
to and from categorical data representations.
"""
import numpy as np
import pandas as pd

from pdcast.decorators.dispatch import dispatch
from pdcast.detect import detect_type


@dispatch("series")
def categorize(
    series: pd.Series,
    levels: list | None = None
) -> pd.Series:
    """Transform a non-categorical series into a categorical series with the
    given levels.

    Parameters
    ----------
    series : pandas.Series
        The series to transform.
    levels : list, optional
        The levels to use for the categorical series.  If this is omitted, then
        levels will be automatically discovered when this method is called.

    Returns
    -------
    pandas.Series
        A categorical series with the same values as the input.

    Notes
    -----
    This function is automatically invoked whenever :func:`cast() <pdcast.cast>`
    is called with a categorical type as a target.  The result will be computed
    at the scalar level and then categorized using this function.

    Examples
    --------
    .. doctest::

        >>> import pandas as pd

        >>> categorize(pd.Series([1, 2, 3]))
        0    1
        1    2
        2    3
        dtype: category
        Categories (3, int64): [1, 2, 3]
    """
    # using a naked CategoricalDtype will automatically infer levels
    if levels is None:
        categorical_type = pd.CategoricalDtype()
    else:
        categorical_type = pd.CategoricalDtype(
            pd.Index(levels, dtype=detect_type(series).dtype)
        )

    return series.astype(categorical_type)


@dispatch("series")
def decategorize(series: pd.Series) -> pd.Series:
    """Transform a categorical series into a non-categorical series.

    Parameters
    ----------
    series : pandas.Series
        The series to transform.

    Returns
    -------
    pandas.Series
        A non-categorical series with the same values as the input.

    Notes
    -----
    This function is automatically invoked whenever :func:`cast() <pdcast.cast>`
    is called on a categorical series.  It is used to convert the series to a
    non-categorical format before casting to the target type.

    Examples
    --------
    .. doctest::

        >>> import pandas as pd

        >>> decategorize(pd.Series([1, 2, 3], dtype="category"))
        0    1
        1    2
        2    3
        dtype: int64
    """
    return series.astype(detect_type(series).wrapped.dtype, copy=False)



"""This module describes an ``@dispatch`` decorator that implements multiple
dispatch based on the inferred type of a function's arguments.
"""
from __future__ import annotations
import inspect
import itertools
import threading
from types import MappingProxyType
from typing import Any, Callable, Hashable, Iterator, Mapping
import warnings

import numpy as np
import pandas as pd

from pdcast.detect import detect_type  # type: ignore
from pdcast.resolve import resolve_type  # type: ignore
from pdcast import types
from pdcast.util.structs import LRUDict
from pdcast.util.type_hints import type_specifier

from .base import Arguments, FunctionDecorator, Signature


# TODO: emit a warning whenever an implementation is replaced.
# -> use a simplefilter when the convert/ module is loaded.


# TODO: None wildcard value?
# -> use registry wildcards instead

# TODO: result is None -> fill with NA?
# -> probably not.  Just return an empty series if filtering.


# TODO: pdcast.cast("today", "datetime") uses python datetimes rather than
# pandas
# -> this because of an errant .larger lookup that made it through the
# refactor.


# TODO: consider empty series during dispatch.
# -> pdcast.cast([None, None, None], int)
# Traceback (most recent call last):
#     ...
# AttributeError: 'NoneType' object has no attribute 'dtype'


######################
####    PUBLIC    ####
######################


def dispatch(
    *args: str,
    drop_na: bool = True,
    fill_na: bool = True,
    rectify: bool = False,
    cache_size: int = 128
) -> DispatchFunc:
    """A decorator that allows a Python function to dispatch to multiple
    implementations based on the type of one or more of its arguments.

    Parameters
    ----------
    *args : str
        The names of the arguments to dispatch on.  These must occur within the
        decorated function's signature, and their order determines the order of
        the keys under which implementations are searched and stored.
    drop_na : bool, default True
        Indicates whether to drop missing values from input vectors before
        forwarding to a dispatched implementation.  If this is set to ``True``,
        then any row that contains one or more missing values will be excluded
        from the dispatch process.  If it is set to ``False``, then missing
        values will be grouped and processed independently, as if they were a
        separate type within the input data.  Optimal performance is achieved
        by setting this to ``True``.
    fill_na : bool, default True
        Indicates whether to replace missing values in the output of a
        dispatched operation.  If this is set to ``True``, then the result will
        be padded with missing values wherever the normalized output index does
        not match that of the input.  This includes any indices that were
        filtered out as a result of ``drop_na=True`` as well as any that were
        dropped within a dispatched implementation itself.  Conversely, if this
        is set to ``False``, then the output will be returned as-is, and no
        missing values will be inserted.
    rectify : bool, default False
        Indicates whether to attempt standardization of mixed-type results.  If
        composite data is encountered and two or more implementations return
        results of different types (but within the same family), then this
        argument determines whether to return them as-is (``False``) or attempt
        to convert them to a common dtype (``True``).  The dtype that is chosen
        will always be the largest of the candidates, as determined by the
        :meth:`< <pdcast.ScalarType.__lt__>` operator.
    cache_size : int, default 128
        Implementation searches are performed every time the decorated function
        is executed, which may be expensive depending on how many
        implementations are being dispatched to and how often the function is
        being called.  To mitigate this,
        :class:`DispatchFuncs <pdcast.dispatch.DispatchFunc>` maintain an LRU
        cache containing the ``n`` most recently requested implementations,
        where ``n`` is equal to the value of this argument.  This cache is
        always checked before performing a full search on the
        :attr:`overloaded <pdcast.DispatchFunc.overloaded>` table itself.

    Returns
    -------
    DispatchFunc
        A cooperative decorator that manages dispatched implementations for the
        decorated callable.

    Raises
    ------
    TypeError
        If the decorated function does not accept the named arguments, or if no
        arguments are given.
    """
    if not args or len(args) == 1 and callable(args[0]):
        raise TypeError("@dispatch requires at least one named argument")

    def decorator(func: Callable) -> DispatchFunc:
        """Convert a callable into a DispatchFunc object."""
        return DispatchFunc(
            func,
            dispatched=args,
            drop_na=drop_na,
            fill_na=fill_na,
            rectify=rectify,
            cache_size=cache_size
        )

    return decorator


#######################
####    PRIVATE    ####
#######################


class DispatchFunc(FunctionDecorator):
    """A wrapper for a function that can dispatch to a collection of virtual
    implementations based on the inferred type of its arguments.

    Parameters
    ----------
    func : Callable
        The decorated function or other callable.
    dispatched : tuple[str, ...]
        The names of the arguments that the function dispatches on.  These
        must occur within the decorated function's signature, and their order
        determines the order of the keys under which implementations are
        searched and stored.
    drop_na : bool
        Indicates whether to drop missing values from input vectors before
        forwarding to a dispatched implementation.  If this is set to ``True``,
        then any row that contains one or more missing values will be excluded
        from the dispatch process.  If it is set to ``False``, then missing
        values will be grouped and processed independently, as if they were a
        separate type within the input data.  Optimal performance is achieved
        by setting this to ``True``.
    fill_na : bool
        Indicates whether to replace missing values in the output of a
        dispatched operation.  If this is set to ``True``, then the result will
        be padded with missing values wherever the normalized output index does
        not match that of the input.  This includes any indices that were
        filtered out as a result of ``drop_na=True`` as well as any that were
        dropped within a dispatched implementation itself.  Conversely, if this
        is set to ``False``, then the output will be returned as-is, and no
        missing values will be inserted.
    rectify : bool
        Indicates whether to attempt standardization of mixed-type results.  If
        composite data is encountered and two or more implementations return
        results of different types (but within the same family), then this
        argument determines whether to return them as-is (``False``) or attempt
        to convert them to a common dtype (``True``).  The dtype that is chosen
        will always be the largest of the candidates, as determined by the
        :meth:`< <pdcast.ScalarType.__lt__>` operator.
    cache_size : int
        Implementation searches are performed every time the decorated function
        is executed, which may be expensive depending on how many
        implementations are being dispatched to and how often the function is
        being called.  To mitigate this,
        :class:`DispatchFuncs <pdcast.dispatch.DispatchFunc>` maintain an LRU
        cache containing the ``n`` most recently requested implementations,
        where ``n`` is equal to the value of this argument.  This cache is
        always checked before performing a full search on the
        :attr:`overloaded <pdcast.DispatchFunc.overloaded>` table itself.

    Examples
    --------
    See the docs for :func:`@dispatch <pdcast.dispatch>` for example usage.
    """

    _reserved = (
        FunctionDecorator._reserved |
        {"_signature", "_flags", "_fill_na", "_rectify"}
    )

    def __init__(
        self,
        func: Callable,
        dispatched: tuple[str, ...],
        drop_na: bool,
        fill_na: bool,
        rectify: bool,
        cache_size: int
    ):
        super().__init__(func=func)
        self._flags = threading.local()  # used to store thread-local state
        self._signature = DispatchSignature(
            func,
            dispatched=dispatched,
            cache_size=cache_size,
            drop_na=drop_na,
        )
        self._fill_na = fill_na
        self._rectify = rectify

    @property
    def dispatched(self) -> tuple[str, ...]:
        """The names of the arguments that this function dispatches on.

        Returns
        -------
        tuple[str, ...]
            The tuple containing the name of every dispatched argument, as
            provided to :func:`@dispatch <pdcast.dispatch>`.

        Notes
        -----
        The order of these names is significant, as it determines the order of
        the keys under which dispatched implementations are stored.  The keys
        themselves are always sorted according to their specificity, but ties
        can still occur if two or more implementations are equally valid for a
        given set of inputs.  For example:

        .. code:: python

            @dispatch("a", "b")
            def foo(a, b):
                ...

            @foo.overload("int", "int[python]")
            def foo1(a, b):
                ...

            @foo.overload("int[python]", "int")
            def foo2(a, b):
                ...

            >>> foo["int[python]", "int[python]"]
            ???

        In this case, it's not immediately clear which implementation we should
        choose.  ``foo1()`` is more specific in its first argument and
        ``foo2()`` in its second.

        In these cases, :func:`@dispatch <pdcast.dispatch>` always **prefers
        implementations** that are **more specific in their first argument**,
        from left to right.  In the case above, this means we always choose
        ``foo2()`` over ``foo1()``.  This behavior can be reversed by changing
        the order of the arguments to :func:`@dispatch() <pdcast.dispatch>`.

        .. code:: python

            @dispatch("b", "a")  # reversed
            def foo(a, b):
                ...

            @foo.overload("int", "int[python]")  # same order (a, b)
            def foo1(a, b):
                ...

            @foo.overload("int[python]", "int")  # same order (a, b)
            def foo2(a, b):
                ...

        We will now always prefer ``foo1()`` over ``foo2()``.

        More complicated priorities can be assigned if a function has 3 or
        more arguments.

        .. note::

            The argument order given in in :func:`@dispatch() <pdcast.dispatch>`
            **does not affect** the signature of
            :meth:`@overload() <pdcast.DispatchFunc.overload>` in any way.  The
            latter always parses its arguments relative to the function that it
            decorates.

            At an implementation level,
            :meth:`@overload() <pdcast.DispatchFunc.overload>` binds its
            arguments directly to the decorated function's signature.  It then
            extracts the dispatched arguments by name, resolves them, and sorts
            them into the order specified by
            :func:`@dispatch() <pdcast.dispatch>`.  All changing this order
            does is reverse the keys that are used to index the
            :class:`DispatchFunc <pdcast.DispatchFunc>`, and thereby consider
            them from right to left (``b`` before ``a``) instead of left to
            right (``a`` before ``b``).

        Examples
        --------
        .. doctest::

            >>> @dispatch("x", "y")
            ... def add(x, y):
            ...     return x + y

            >>> add.dispatched
            ('x', 'y')
        """
        return self._signature.dispatched

    @property
    def overloaded(self) -> Mapping[tuple[types.VectorType, ...], Callable]:
        """A map connecting :doc:`types </content/types/types>` to their
        :meth:`overloaded <pdcast.DispatchFunc.overload>` implementations.

        Returns
        -------
        Mapping[tuple[types.VectorType, ...], Callable]
            A read-only dictionary mapping types to their associated
            implementations.  The keys are tuples that are sorted into the same
            order as the :attr:`dispatched <pdcast.DispatchFunc.dispatched>`
            arguments.

        Notes
        -----
        The returned mapping is `topologically sorted
        <https://en.wikipedia.org/wiki/Topological_sorting>`_ according to
        specificity.  Iterating through the map equates to searching it from
        most to least specific.

        Examples
        --------
        .. doctest::

            >>> @dispatch("bar")
            ... def foo(bar):
            ...     print("base")
            ...     return bar

            >>> @foo.overload("int")  # least specific
            ... def integer_foo(bar):
            ...     print("integer")
            ...     return bar

            >>> @foo.overload("int64[numpy]")  # most specific
            ... def numpy_int64_foo(bar):
            ...     print("int64[numpy]")
            ...     return bar

            >>> @foo.overload("int64")
            ... def int64_foo(bar):
            ...     print("int64")
            ...     return bar

            >>> for key, func in foo.overloaded.items():
            ...     print(f"{key}: {func}")
            (NumpyInt64Type(),): <function numpy_int64_foo at ...>
            (Int64Type(),): <function int64_foo at ...>
            (IntegerType(),): <function integer_foo at ...>
        """
        # sort the dispatch map if it hasn't been sorted already.  Usually,
        # this is done lazily the first time the function is called.
        if not self._signature.ordered:
            self._signature.sort()

        return self._signature.dispatch_map

    def overload(
        self,
        *args: type_specifier,
        **kwargs: type_specifier
    ) -> Callable[..., Any]:
        """A decorator that transforms a function into a dispatched
        implementation for this :class:`DispatchFunc <pdcast.DispatchFunc>`.

        Parameters
        ----------
        *args, **kwargs
            The type(s) to dispatch on.  These are provided as positional and
            keyword arguments that are bound to the decorated function's
            signature.  They have the same semantics as if the function
            were being invoked directly, and can be in any form recognized by
            :func:`resolve_type() <pdcast.resolve_type>`.

        Returns
        -------
        Callable
            The same function as was decorated.  This is not changed in any
            way.

        Raises
        ------
        TypeError
            If the decorated function does not accept the arguments named in
            :func:`@dispatch() <pdcast.dispatch>` (order doesn't matter).

        Notes
        -----
        This decorator works just like the :meth:`register` method of
        :func:`singledispatch <python:functools.singledispatch>` functions,
        except that it does not interact with type annotations in any way.

        Instead, types are declared naturally through the decorator itself and
        bound to the function as if it were being invoked.  They can be
        supplied as either positional or keyword arguments, and can be in any
        format recognized by :func:`resolve_type() <pdcast.resolve_type>`.

        Examples
        --------
        See the :ref:`API docs <dispatch.dispatched>` for example usage.
        """

        def implementation(func: Callable) -> Callable:
            """Attach a dispatched implementation to the DispatchFunc with the
            associated types.
            """
            signature = Signature(func)

            # bind *args, **kwargs to the decorated function
            try:
                bound = signature(*args, **kwargs)
            except TypeError as err:
                func_name = f"'{func.__module__}.{func.__qualname__}()'"
                reconstructed = [repr(value) for value in args]
                reconstructed.extend(
                    f"{name}={repr(value)}" for name, value in kwargs.items()
                )
                err_msg = (
                    f"invalid signature for {func_name}: "
                    f"({', '.join(reconstructed)})"
                )
                raise TypeError(err_msg) from err

            # translate bound arguments into this signature's order
            bound.apply_defaults()
            key = self._signature.dispatch_key(*bound.args, **bound.kwargs)

            # register every combination of types
            for path in self._signature.cartesian_product(key):
                self._signature[path] = func

            return func

        return implementation

    def fallback(self, *args: Any, **kwargs: Any) -> Any:
        """A reference to the base implementation of the dispatch function.

        Parameters
        ----------
        *args, **kwargs
            Positional and keyword arguments to supply to the base
            implementation.  These are passed through directly to the base
            function.

        Returns
        -------
        Any
            The return value of the base implementation.

        Examples
        --------
        This allows direct access to the base implementation of a
        :class:`DispatchFunc <pdcast.DispatchFunc>`, bypassing the dispatch
        mechanism entirely.

        .. doctest::

            >>> @dispatch("bar")
            ... def foo(bar):
            ...     print("base")
            ...     return bar

            >>> @foo.overload("int")
            ... def integer_foo(bar):
            ...     print("integer")
            ...     return bar

            >>> foo(1)
            integer
            1
            >>> foo.fallback(1)
            base
            1
        """
        return self.__wrapped__(*args, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Execute the :class:`DispatchFunc <pdcast.DispatchFunc>`, searching
        for an overloaded implementation.

        Parameters
        ----------
        *args, **kwargs
            Positional and keyword arguments to supply to the virtual
            implementation.  The
            :attr:`dispatched <pdcast.DispatchFunc.dispatched>` arguments are
            automatically extracted from these and analyzed to determine the
            most specific implementation to use.  They may also include
            vectors, which are normalized and grouped according to their
            inferred type.

        Returns
        -------
        Any
            The return value of the dispatched implementation(s).  If any of
            the arguments were vectorized, then this value will be analyzed for
            aggregations, transformations, and filtrations.  See the notes
            below for more information.

        See Also
        --------
        DispatchFunc.__getitem__ :
            Access a specific implementation of the dispatch function.

        Notes
        -----
        :func:`@dispatch() <pdcast.dispatch>` automatically intercepts any
        vectors that are supplied to the
        :attr:`dispatched <pdcast.DispatchFunc.dispatched>` arguments and
        converts them into properly-formatted
        :class:`pandas.Series <pandas.Series>` objects.  These are always
        normalized and labeled with an appropriate ``dtype`` before being
        passed into the function itself.  Scalars are always passed through
        as-is.

        The steps that are taken to normalize input vectors are as follows:

            #.  Convert each vector into a
                :class:`pandas.Series <pandas.Series>` object if it is not one
                already.  If a vector does not have an explicit ``dtype``, then
                it is treated as a ``dtype: object`` series.
            #.  Bind each :class:`Series <pandas.Series>` into a shared
                :class:`DataFrame <pandas.DataFrame>` with a column for each
                argument.
            #.  Normalize the frame's collective index to be a
                :class:`RangeIndex <pandas.RangeIndex>` spanning the length of
                each vector.  If the vectors had a custom index, then it is
                stored until the end of the dispatch process.
            #.  If ``drop_na=True``, drop any row that contains a missing value
                in one or more columns.  This creates gaps in the index, which
                are used to identify the locations of missing values later on
                in the dispatch process.  If ``drop_na=False``, then we skip
                this step and treat the missing values as a separate group
                instead.
            #.  Detect the type of each vector and label it accordingly.  This
                allows :func:`detect_type() <pdcast.detect_type>` to infer each
                vector's type in constant time within the dispatched context.
                If any of the vectors contain data of mixed type, then the
                whole frame is split into groups based on the observed type at
                every index.  Each group is then passed through individually
                and combined afterwards according to its index.

        The dispatched implementation is then invoked with the normalized
        arguments, and its return value is analyzed for aggregations,
        transformations, and filtrations.  These are determined based on the
        return type.

            *   A pandas :class:`Series <pandas.Series>` or
                :class:`DataFrame <pandas.DataFrame>` signifies a
                transformation.  If ``fill_na=True``, then upon exiting the
                dispatched context, any missing indices will be replaced with
                the appropriate :attr:`na_value <pdcast.ScalarType.na_value>`.
                Otherwise, the concatenated output will be used as-is.  If the
                arguments defined a custom index, then it will be replaced at
                this point, and the output ``dtype`` will always correspond to
                the type of the resulting elements.
            *   :data:`None <python:None>` signifies a filtration.  In most
                cases, this will be passed through as-is.  However, if the
                input includes a vector of mixed type, then the group will be
                removed from the output :class:`DataFrame <pdcast.DataFrame>`.
            *   Anything else signifies an aggreggation.  If the input data are
                homogenous, then the result will be returned immediately.
                Otherwise, the result will be a
                :class:`DataFrame <pandas.DataFrame>` with c0olumns for each of
                the :attr:`dispatched <pdcast.DispatchFunc.dispatched>`
                arguments.  These contain the observed type of the argument for
                each group, and the final column contains the result of the
                computation for that group.

        These behaviors can be customized using the keyword arguments to
        :func:`@dispatch() <pdcast.dispatch>`.  See the
        :ref:`documentation <dispatch.missing_values>` for more details.

        A full :ref:`activity diagram <dispatch.activity_diagram>` describing
        the dispatch process is also provided for clarity.

        Examples
        --------
        See the :ref:`API docs <dispatch>` for example usage.
        """
        strategy: DispatchStrategy

        # bind arguments
        bound = self._signature(*args, **kwargs)

        # fastpath: if calling from a recursive context, skip normalization
        if getattr(self._flags, "recursive", False):
            strategy = HomogenousDispatch(
                self,
                arguments=bound,
                fill_na=self._fill_na,
            )
            return strategy.execute()  # do not finalize

        # normalize arguments
        bound.normalize()

        # choose strategy
        if not bound.is_composite:
            strategy = HomogenousDispatch(
                self,
                arguments=bound,
                fill_na=self._fill_na,
            )
        else:
            strategy = CompositeDispatch(
                self,
                arguments=bound,
                fill_na=self._fill_na,
                rectify=self._rectify,
            )

        # execute strategy
        self._flags.recursive = True
        try:
            return strategy.finalize(strategy.execute())
        finally:
            self._flags.recursive = False

    def __getitem__(
        self,
        key: type_specifier | tuple[type_specifier, ...]
    ) -> Callable:
        """Get the dispatched implementation for a particular combination of
        argument types.

        Parameters
        ----------
        key : type_specifier | tuple[type_specifier, ...]
            The input types associated with a particular implementation.
            Multiple types can be given by separating them with commas, and
            they must be given in the same order as the arguments supplied to
            :func:`@dispatch() <pdcast.dispatch>`.

        Returns
        -------
        Callable
            The implementation that will be chosen when the function is invoked
            with the given type(s).  If no specific implementation is found
            for the associated types, then a reference to the default
            implementation is returned instead.

        See Also
        --------
        DispatchFunc.__call__ :
            Call the :class:`DispatchFunc <pdcast.DispatchFunc>` with the
            associated implementation.
        DispatchFunc.__delitem__ :
            Remove a dispatched implementation from the pool.

        Notes
        -----
        This always returns the same implementation that is chosen when the
        dispatch mechanism is executed.  In fact, the mechanism simply calls
        this method to retrieve the implementation in the first place.

        Examples
        --------
        Types must be supplied in the same order as the
        :attr:`dispatched <pdcast.DispatchFunc.dispatched>` arguments.

        .. doctest::

            >>> cast[int, int]
            <function integer_to_integer at ...>
            >>> cast[float, bool]
            <function float_to_boolean at ...>
            >>> cast[int, "datetime[pandas, US/Pacific]"]
            <function integer_to_pandas_timestamp at ...>
        """
        try:
            return self._signature[key]
        except KeyError:
            return self.__wrapped__

    def __delitem__(
        self,
        key: type_specifier | tuple[type_specifier, ...]
    ) -> None:
        """Remove a dispatched implementation from the pool.

        Parameters
        ----------
        key : type_specifier | tuple[type_specifier, ...]
            The input types associated with a particular implementation.
            Multiple types can be given by separating them with commas, and
            they must be given in the same order as the arguments supplied to
            :func:`@dispatch() <pdcast.dispatch>`.

        Raises
        ------
        KeyError
            If the given type(s) are not associated with any
            :meth:`overloaded <pdcast.DispatchFunc.overload>` implementation.

        See Also
        --------
        DispatchFunc.__call__ :
            Call the :class:`DispatchFunc <pdcast.DispatchFunc>` with the
            associated implementation.
        DispatchFunc.__getitem__ :
            Get the dispatched implementation associated with the given
            type(s).

        Notes
        -----
        Just like :meth:`__getitem__() <pdcast.DispatchFunc.__getitem__>`, this
        method always removes the same implementation that is chosen when the
        dispatch mechanism is actually executed.  This means that it does not
        require an exact match for the input types.  Instead, it will always
        remove the most specific implementation that matches the given types.

        The default implementation cannot be removed.

        Examples
        --------
        Types must be supplied in the same order as the
        :attr:`dispatched <pdcast.DispatchFunc.dispatched>` arguments.

        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> @foo.overload("int", "int")
            ... def integer_foo(x, y):
            ...     return "integer"

            >>> foo(1, 2)
            'integer'
            >>> foo[int, int]
            <function integer_foo at ...>
            >>> del foo[int, int]
            >>> foo[int, int]
            <function foo at ...>
            >>> foo(1, 2)
            'default'
        """
        del self._signature[key]


class DispatchSignature(Signature):
    """A wrapper around an :class:`inspect.Signature <python:inspect.Signature>`
    object that serves as a factory for
    :class:`DispatchArguments <pdcast.DispatchArguments>`.

    Parameters
    ----------
    func : Callable
        The function whose signature will be wrapped.
    dispatched : tuple[str, ...]
        The names of the arguments that will be used to select a dispatched
        implementation.  These must be a subset of the function's parameters,
        and they can be defined in any order.  The order dictates the
        precedence of the dispatched implementations.
    cache_size : int
        The maximum number of dispatched implementations to cache.  This
        short-circuits the dispatch mechanism for previously used data types.
    drop_na : bool
        Indicates whether to drop missing values from the arguments during
        dispatch.

    Notes
    -----
    One of these objects is associated with every
    :class:`DispatchFunc <pdcast.DispatchFunc>` that ``pdcast`` creates.  They
    are responsible for parsing the arguments that are supplied to a function
    and mapping them to the correct implementation for their detected types.

    To this end, every :class:`DispatchSignature <pdcast.DispatchSignature>`
    maintains an ordered dictionary of all the dispatched implementations that
    are associated with the parent function.  Whenever the function is invoked,
    a key will be generated to index this dictionary, returning the most
    specific implementation that matches the input types.
    """

    def __init__(
        self,
        func: Callable,
        dispatched: tuple[str, ...],
        cache_size: int,
        drop_na: bool
    ):
        self._dispatch_map: dict[tuple[types.VectorType, ...], Callable]

        super().__init__(func=func)
        missing = [arg for arg in dispatched if arg not in self.parameter_map]
        if missing:
            raise TypeError(f"argument not recognized: {missing}")

        self.dispatched = dispatched
        self._dispatch_map = {}
        self.cache = LRUDict(maxsize=cache_size)
        self.ordered = False
        self.drop_na = drop_na

    @property
    def dispatch_map(self) -> Mapping[tuple[types.VectorType, ...], Callable]:
        """A map containing all the dispatched implementations being handled by
        this :class:`DispatchSignature <pdcast.DispatchSignature>`.

        Returns
        -------
        MappingProxyType
            A read-only mapping from
            :meth:`dispatch_keys <pdcast.DispatchSignature.dispatch_key>` to
            their respective
            :meth:`implementations <pdcast.DispatchFunc.overload>`.

        See Also
        --------
        DispatchSignature.__getitem__ :
            Index the dispatch map for a particular combination of types.
        DispatchSignature.__setitem__ :
            Set an implementation within the dispatch map.
        DispatchSignature.__delitem__ :
            Delete an implementation from the map.

        Notes
        -----
        The map is not guaranteed to be in the correct order.  To sort it, use
        the :meth:`sort() <pdcast.DispatchSignature.sort>` method.

        Examples
        --------
        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> @foo.overload("int", "int")
            ... def integer_foo(x, y):
            ...     return "integer"

            >>> foo._signature.dispatch_map
            mappingproxy({(IntegerType(), IntegerType()): <function integer_foo at ...>})
        """
        return MappingProxyType(self._dispatch_map)

    def sort(self) -> None:
        """Sort the dictionary into topological order, with the most specific
        keys first.

        Notes
        -----
        This method is called automatically whenever the dispatch map is
        searched without an exact match.  It can also be called manually to
        sort the map on demand.

        Examples
        --------
        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> @foo.overload("int", "int")
            ... def integer_foo(x, y):
            ...     return "integer"

            >>> @foo.overload("int32", "int32")  # int32 is more specific than int
            ... def int32_foo(x, y):
            ...     return "int32"

            >>> foo._signature.dispatch_map
            mappingproxy({(IntegerType(), IntegerType()): <function integer_foo at ...>, (Int32Type(), Int32Type()): <function int32_foo at ..>})
            >>> foo._signature.sort()
            >>> foo._signature.dispatch_map
            mappingproxy({(Int32Type(), Int32Type()): <function int32_foo at ...>, (IntegerType(), IntegerType()): <function integer_foo at ...>})
        """
        edges: dict[
            tuple[types.VectorType, ...],
            set[tuple[types.VectorType, ...]]
        ]

        # draw edges between keys according to their specificity
        edges = {key: set() for key in self._dispatch_map}
        for key1 in self._dispatch_map:
            for key2 in self._dispatch_map:
                if edge(key1, key2):
                    edges[key1].add(key2)

        # sort according to edges
        for key in topological_sort(edges):
            # NOTE: (Python 3.6+) equivalent to OrderedDict.move_to_end()
            item = self._dispatch_map.pop(key)
            self._dispatch_map[key] = item

        self.ordered = True

    # pylint: disable=no-self-argument
    def dispatch_key(__self, *args, **kwargs) -> tuple[types.Type, ...]:
        """Form a dispatch key from the provided arguments.

        Parameters
        ----------
        *args, **kwargs
            Arbitrary positional and/or keyword arguments to bind to this
            signature.  These can be provided in any format recognized by
            :func:`resolve_type() <pdcast.resolve_type>`.  Any that are
            marked as dispatched arguments will be extracted and resolved in
            the same order that they were defined in this signature's
            constructor.

        Returns
        -------
        tuple[types.Type, ...]
            A tuple of resolved types that can be used as a key to the
            signature's
            :attr:`dispatch_map <pdcast.DispatchSignature.dispatch_map>`.  The
            order of this tuple will always match the order of the dispatched
            arguments that was specified when this signature was constructed.

        Raises
        ------
        KeyError
            If any of the dispatched arguments are not provided.

        See Also
        --------
        DispatchSignature.cartesian_product :
            Expand composite keys into a cartesian product of all possible
            combinations.

        Notes
        -----
        :class:`composite <pdcast.CompositeType>` specifiers will be resolved
        by this method as normal.  However, since they are not hashable, they
        cannot be stored in :attr:`dispatch_map <pdcast.dispatch_map>`
        directly.  Instead they must be expanded using
        :meth:`DispatchSignature.cartesian_product() <pdcast.DispatchSignature.cartesian_product>`
        and stored independently.

        Examples
        --------
        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> foo._signature.dispatch_key("int", "int")
            (IntegerType(), IntegerType())
            >>> foo._signature.dispatch_key(x="bool", y="float")
            (BooleanType(), FloatType())
            >>> foo._signature.dispatch_key(y="decimal", x="complex")
            (ComplexType(), DecimalType())
        """
        bound = __self.sig.bind_partial(*args, **kwargs)
        bound.apply_defaults()

        return tuple(
            resolve_type(bound.arguments[x]) for x in __self.dispatched
        )

    def cartesian_product(
        self,
        key: tuple[types.Type, ...]
    ) -> Iterator[tuple[types.VectorType, ...]]:
        """Convert a composite
        :meth:`dispatch key <pdcast.DispatchSignature.dispatch_key>` into a
        Cartesian product containing all possible combinations.

        Parameters
        ----------
        key : tuple[types.Type, ...]
            A :meth:`dispatch key <pdcast.DispatchSignature.dispatch_key>` to
            parse.

        Returns
        -------
        Iterator[tuple[types.VectorType, ...]]
            A generator that yields non-composite dispatch keys representing
            every combination of the input types.

        See Also
        --------
        DispatchSignature.dispatch_key :
            Create a dispatch key by binding to the signature.

        Examples
        --------
        This method expands dispatch keys into a sequence of non-composite
        types.

        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> key = foo._signature.dispatch_key("int, float, complex", "int")
            >>> key
            (CompositeType({int, float, complex}), IntegerType())
            >>> list(foo._signature.cartesian_product(key))
            [(IntegerType(), IntegerType()), (FloatType(), IntegerType()), (ComplexType(), IntegerType())]

        These types can then be used to index the
        :attr:`dispatch_map <pdcast.DispatchSignature.dispatch_map>` as normal.
        """
        # convert keys into CompositeTypes
        key = tuple(resolve_type([x]) for x in key)
        for path in itertools.product(*key):
            yield path

    # pylint: disable=no-self-argument
    def __call__(__self, *args, **kwargs) -> DispatchArguments:
        """Bind the arguments to this signature and return a corresponding
        :class:`DispatchArguments <pdcast.DispatchArguments>` object.

        Parameters
        ----------
        *args, **kwargs
            Arbitrary positional and/or keyword arguments to bind to this
            signature.

        Returns
        -------
        DispatchArguments
            A new :class:`DispatchArguments <pdcast.DispatchArguments>` object,
            that encapsulates the bound arguments.  These are used as context
            objects for :class:`DispatchStrategies <pdcast.DispatchStrategy>`,
            which manipulate them to perform the actual dispatch.

        Notes
        -----
        This is always called on the input to
        :meth:`DispatchFunc.__call__() <pdcast.DispatchFunc.__call__>`.
        """
        bound = __self.sig.bind_partial(*args, **kwargs)
        bound.apply_defaults()

        return DispatchArguments(
            bound=bound,
            signature=__self,
            drop_na=__self.drop_na
        )

    def __getitem__(
        self,
        key: type_specifier | tuple[type_specifier]
    ) -> Callable:
        """Search the :class:`DispatchSignature <pdcast.DispatchSignature>` for
        a particular implementation.

        Parameters
        ----------
        key : type_specifier | tuple[type_specifier]
            One or more types to search for.  These are passed directly to the
            :meth:`dispatch_key() <pdcast.DispatchSignature.dispatch_key>`
            method, and multiple types can be specified by separating them with
            commas.

        Returns
        -------
        Callable
            The implementation that matches the provided key.  This is found by
            searching the
            :attr:`dispatch_map <pdcast.DispatchSignature.dispatch_map>` for
            the most specific key that fully contains the input types.

        Raises
        ------
        KeyError
            If no implementation matches the provided key.

        Examples
        --------
        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> @foo.overload("int", "int")
            ... def integer_foo(x, y):
            ...     return "integer"

            >>> foo._signature["int", "int"]
            <function integer_foo at ...>
            >>> foo._signature["int32", "int32"]  # subtypes of int
            <function integer_foo at ...>
            >>> foo._signature["int, float"]  # no match for float
            Traceback (most recent call last):
                ...
            KeyError: ('int', 'float')
        """
        if not isinstance(key, tuple):
            key = (key,)

        # resolve key
        # TODO: dispatched arguments might be positional-only
        key = self.dispatch_key(**dict(zip(self.dispatched, key)))

        # trivial case: key has exact match
        if key in self._dispatch_map:
            return self._dispatch_map[key]

        # check for cached result
        if key in self.cache:
            return self.cache[key]

        # sort map
        if not self.ordered:
            self.cache.clear()
            self.sort()

        # search for first (sorted) match that fully contains key
        for dispatch_key, implementation in self._dispatch_map.items():
            if all(y.contains(x) for x, y in zip(key, dispatch_key)):
                self.cache[key] = implementation
                return implementation

        # no match found
        raise KeyError(tuple(str(x) for x in key))

    def __setitem__(
        self,
        key: type_specifier | tuple[type_specifier, ...],
        value: Callable
    ) -> None:
        """Register an overloaded implementation with the
        :class:`DispatchSignature <pdcast.DispatchSignature>`.

        Parameters
        ----------
        key : type_specifier | tuple[type_specifier, ...]
            The key to register the implementation under.  This is passed
            directly to the
            :meth:`dispatch_key() <pdcast.DispatchSignature.dispatch_key>`
            method, and multiple types can be specified by separating them with
            commas.  They must always be specified in the same order as the
            dispatched arguments that were given to this signature.
        value : Callable
            The implementation to register.  This must be a callable object
            that accepts the same arguments as the signature.  It can include
            additional arguments, but it must at least implement this
            signature's interface.

        Raises
        ------
        TypeError
            If the implementation is not callable, or if it has an incompatible
            signature.

        Warns
        -----
        UserWarning
            If the provided key overwrites a previous implementation.

        Notes
        -----
        This is automatically called by the
        :meth:`@overload() <pdcast.DispatchFunc.overload>` decorator.

        Examples
        --------
        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> def integer_foo(x, y):
            ...     return "integer"

            >>> foo._signature["int", "int"]
            <function foo at ...>
            >>> foo._signature["int", "int"] = integer_foo
            >>> foo._signature["int", "int"]
            <function integer_foo at ...>
        """
        if not isinstance(key, tuple):
            key = (key,)

        # resolve key
        key = self.dispatch_key(**dict(zip(self.dispatched, key)))

        # verify implementation is callable
        if not callable(value):
            raise TypeError(f"implementation must be callable: {repr(value)}")

        # verify implementation has compatible signature
        signature = Signature(value)
        if not self.compatible(signature):
            self_str = self.reconstruct(
                defaults=False,
                annotations=False,
                return_annotation=False
            )
            other_str = signature.reconstruct(
                defaults=False,
                annotations=False,
                return_annotation=False
            )
            raise TypeError(
                f"signatures are not compatible: '{other_str}' is not an "
                f"extension of '{self_str}'"
            )

        # warn if overwriting a previous key
        if key in self._dispatch_map:
            warn_msg = (
                f"Replacing '{self.dispatch_map[key].__qualname__}()' "
                f"with '{value.__qualname__}()' for signature "
                f"{tuple(str(x) for x in key)}"
            )
            # warnings.warn(warn_msg, UserWarning, stacklevel=2)

        # insert into dispatch map
        self._dispatch_map[key] = value
        self.ordered = False

    def __delitem__(
        self,
        key: type_specifier | tuple[type_specifier, ...]
    ) -> None:
        """Remove an overloaded implementation from the
        :class:`DispatchSignature <pdcast.DispatchSignature>`.

        Parameters
        ----------
        key : type_specifier | tuple[type_specifier, ...]
            The key to remove.  This is passed directly to the
            :meth:`dispatch_key() <pdcast.DispatchSignature.dispatch_key>`
            method, and multiple types can be specified by separating them with
            commas.  They must always be specified in the same order as the
            dispatched arguments that were given to this signature.

        Raises
        ------
        KeyError
            If no implementation matches the provided key.

        Examples
        --------
        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> @foo.overload("int", "int")
            ... def integer_foo(x, y):
            ...     return "integer"

            >>> foo._signature["int", "int"]
            <function integer_foo at ...>
            >>> del foo._signature["int", "int"]
            >>> foo._signature["int", "int"]
            <function foo at ...>
        """
        if not isinstance(key, tuple):
            key = (key,)

        # resolve key
        key = self.dispatch_key(**dict(zip(self.dispatched, key)))

        # trivial case: key has exact match
        if key in self._dispatch_map:
            del self._dispatch_map[key]
            return None

        # sort map
        if not self.ordered:
            self.sort()

        # search for first (sorted) match that fully contains key
        # pylint: disable=consider-using-dict-items
        for dispatch_key in self._dispatch_map:
            if all(y.contains(x) for x, y in zip(key, dispatch_key)):
                del self._dispatch_map[dispatch_key]
                self.cache.clear()
                return None

        # no match found
        raise KeyError(tuple(str(x) for x in key))

    def __contains__(
        self,
        key: type_specifier | tuple[type_specifier, ...]
    ) -> bool:
        """Check if a particular implementation is present in the
        :class:`DispatchSignature <pdcast.DispatchSignature>`.

        Parameters
        ----------
        key : type_specifier | tuple[type_specifier, ...]
            The key to check for.  This is passed directly to the
            :meth:`dispatch_key() <pdcast.DispatchSignature.dispatch_key>`
            method, and multiple types can be specified by separating them with
            commas.  They must always be specified in the same order as the
            dispatched arguments that were given to this signature.

        Returns
        -------
        bool
            ``True`` if the key has a matching implementation.  ``False``
            otherwise.

        Examples
        --------
        .. doctest::

            >>> @dispatch("x", "y")
            ... def foo(x, y):
            ...     return "default"

            >>> @foo.overload("int", "int")
            ... def integer_foo(x, y):
            ...     return "integer"

            >>> ("int", "int") in foo._signature
            True
            >>> ("int", "str") in foo._signature
            False
        """
        try:
            self.__getitem__(key)
            return True
        except KeyError:
            return False


class DispatchArguments(Arguments):
    """A wrapper for an `inspect.BoundArguments` object with extra context for
    dispatched arguments and their detected types.

    Parameters
    ----------
    bound : inspect.BoundArguments
        The bound arguments to wrap.  These are produced by binding an
        :class:`inspect.Signature <python:inspect.Signature>` object to a
        collection of positional and/or keyword arguments.
    signature : DispatchSignature
        A reference to the :class:`DispatchSignature <pdcast.DispatchSignature>`
        that was used to produce the bound arguments.
    drop_na : bool
        A flag indicating whether to drop missing values during data
        normalization/type detection.
    detected : dict[str, types.Type], optional
        A dictionary mapping argument names to their detected types.  This is
        usually generated at runtime, but it can be provided manually as an
        optimization to short-circuit the detection process and limit calls
        to :func:`detect_type <pdcast.detect_type>`.
    normalized : bool, optional
        A flag indicating whether the arguments have been normalized.  If this
        is set to ``True``, then the
        :meth:`DispatchArguments.normalize <pdcast.DispatchArguments.normalize>`
        method will always raise an error.

    Notes
    -----
    These are created by :class:`DispatchSignatures <pdcast.DispatchSignature>`
    and serve as context objects for ``DispatchStrategies``, which
    control the execution of a :class:`DispatchFunc <pdcast.DispatchFunc>`.  By
    encapsulating the arguments in a separate object, we can cache certain
    operations and minimize the work that needs to be done for each invocation.
    """

    def __init__(
        self,
        bound: inspect.BoundArguments,
        signature: DispatchSignature,
        drop_na: bool,
        detected: dict[str, types.Type] | None = None,
        normalized: bool = False
    ):
        super().__init__(bound=bound, signature=signature)
        self.drop_na = drop_na

        # extract dispatched arguments
        self.dispatched: dict[str, Any] = {
            arg: bound.arguments[arg] for arg in self.signature.dispatched
        }

        # cached in .types
        self._types = detected

        # cached in .normalize()
        self.normalized = normalized
        self.series_names: dict[str, Hashable] = {}
        self.frame: pd.DataFrame | None = None
        self.original_index: pd.Index | None = None
        self.hasnans: bool | None = None

    @property
    def types(self) -> dict[str, types.Type]:
        """The detected type of every :func:`dispatched <pdcast.dispatch>`
        argument.

        Returns
        -------
        dict[str, types.Type]
            A dictionary mapping argument names to their detected types.  This
            is computed once and then cached for the duration of the dispatch
            process.

        Notes
        -----
        The result is always sorted in the same order as the arguments that
        were provided to the :func:`@dispatch <pdcast.dispatch>` decorator.

        The values may be :class:`CompositeTypes <pdcast.CompositeType>`, in
        which case a ``CompositeDispatch`` strategy will be chosen when the
        function is executed.
        """
        # cache result
        if self._types is None:
            self._types = {
                arg: detect_type(val, drop_na=self.drop_na)
                for arg, val in self.dispatched.items()
            }
        return self._types

    @property
    def key(self) -> tuple[types.Type, ...]:
        """Form a key that can be used to index a
        :class:`DispatchFunc <pdcast.DispatchFunc>` in search of a particular
        implementation.

        Returns
        -------
        tuple[types.Type, ...]
            A tuple containing one or more types corresponding to the values of
            the :attr:`DispatchArguments.types <pdcast.DispatchArguments.types>`
            table.

        Notes
        -----
        Keys such as these are used to store and retrieve implementations from
        a :class:`DispatchSignature <pdcast.DispatchSignature>` object.  They
        may contain :class:`CompositeTypes <pdcast.CompositeType>`, in which
        case a ``CompositeDispatch`` strategy will be chosen when the function
        is executed.
        """
        return tuple(self.types.values())

    @property
    def is_composite(self) -> bool:
        """Check if the dispatched arguments are composite.

        Returns
        -------
        bool
            ``True`` if composite input was supplied to any of the function's
            :func:`dispatched <pdcast.dispatch>` arguments.  ``False``
            otherwise.

        Notes
        -----
        This is used to determine whether a ``CompositeDispatch`` strategy
        should be applied when the function is executed.
        """
        return any(isinstance(typ, types.CompositeType) for typ in self.key)

    @property
    def groups(self) -> Iterator[DispatchArguments]:
        """Split input vectors into groups based on their inferred type.

        Yields
        ------
        DispatchArguments
            A separate :class:`DispatchArguments <pdcast.DispatchArguments>`
            context for each group.  Each group will be dispatched
            independently when the function is executed.

        Notes
        -----
        This is only accessed if :attr:`DispatchArguments.is_composite` is set
        to ``True``.

        ``CompositeDispatch`` is achieved by iterating over these groups and
        applying a ``HomogenousDispatch`` strategy to each one independently.
        The results are then collected and combined into a single result.
        """
        assert self.frame is not None

        if not self.is_composite:
            raise RuntimeError("DispatchArguments are not composite")

        # generate type frame using the inferred indices of each vector
        type_frame = pd.DataFrame({
            arg: getattr(typ, "index", typ) for arg, typ in self.types.items()
        })

        # groupby() to get the frame indices of each group
        groupby = type_frame.groupby(list(type_frame.columns), sort=False)
        del type_frame  # free memory

        # extract groups from frame one by one
        for key, indices in groupby.groups.items():
            group = self.frame.iloc[indices]

            # bind argument names to group key
            if not isinstance(key, tuple):
                key = (key,)
            detected = dict(zip(self.dispatched, key))

            # split group into vectors and rectify their dtypes
            vectors = self._extract_vectors(group)
            vectors = self._standardize_dtype(vectors, detected)

            # generate a separate BoundArguments instance for each group
            bound = type(self.bound)(
                arguments=self.bound.arguments | vectors,
                signature=self.bound.signature
            )

            # convert to DispatchArguments and yield to CompositDispatch
            yield DispatchArguments(
                bound=bound,
                signature=self.signature,
                drop_na=False,  # already handled by groupby()
                detected=detected,  # short-circuit type detection
                normalized=True  # block normalize() calls
            )

    def normalize(self) -> None:
        """Extract vectors from dispatched arguments and normalize them.

        Raises
        ------
        RuntimeError
            If this method is called more than once, or if it is called out of
            order with respect to
            :attr:`DispatchArguments.types <pdcast.DispatchArguments.types>`.
            This is to ensure that no performance regressions crop up due to
            future changes.

        Notes
        -----
        This method works by binding the dispatched vectors into a shared
        DataFrame, and then normalizing that DataFrame.  The normalized vectors
        are then split back into separate arguments, replacing their original
        equivalents.  For performance reasons, this is only called once.
        """
        # ensure normalize() is only called once
        if self.normalized:
            raise RuntimeError("DispatchArguments have already been normalized")

        # ensure types have not yet been detected
        if self._types is not None:
            raise RuntimeError("types were inferred before normalizing")

        # extract vectors from dispatched args and bind them to a DataFrame
        self._build_frame()

        # convert index to RangeIndex
        self._replace_index()

        # drop missing values
        if self.drop_na:
            self._dropna()

        # split DataFrame into vectors and replace original args
        vectors = self._extract_vectors(self.frame)  # type: ignore
        self.bound.arguments |= vectors

        # rectify dtypes
        if not self.is_composite:  # NOTE: implicitly detects type of each arg
            self.bound.arguments |= self._standardize_dtype(vectors, self.types)

        # mark as normalized
        self.normalized = True

    def _build_frame(self) -> None:
        """Extract vectors from dispatched arguments and bind them into a
        shared DataFrame.

        The resulting frame is accessible under ``DispatchArguments.frame``.
        """
        # extract vectors
        vectors: dict[str, pd.Series | np.ndarray] = {}
        for arg, value in self.dispatched.items():
            if isinstance(value, pd.Series):
                vectors[arg] = value
                self.series_names[arg] = value.name
            elif isinstance(value, np.ndarray):
                vectors[arg] = value
            else:
                value = np.asarray(value, dtype=object)
                if value.shape:
                    vectors[arg] = value

        # bind vectors into DataFrame
        self.frame = pd.DataFrame(vectors)

        # NOTE: the DataFrame constructor will happily accept vectors with
        # misaligned indices, filling any absent values with NaNs.  Since this
        # is a likely source of bugs, we always warn the user when it happens.

        # warn if indices are misaligned
        original_length = self.frame.shape[0]
        if any(vec.shape[0] != original_length for vec in vectors.values()):
            warn_msg = (
                f"{self.__qualname__}() - vectors have misaligned indices"
            )
            warnings.warn(warn_msg, UserWarning, stacklevel=3)

    def _replace_index(self) -> None:
        """Replace the frame index with a unique
        :class:`RangeIndex <pandas.RangeIndex>` that has no duplicate values.

        The previous index is retained under
        ``DispatchArguments.original_index``.
        """
        assert self.frame is not None

        self.original_index = self.frame.index
        if not isinstance(self.original_index, pd.RangeIndex):
            self.frame.index = pd.RangeIndex(0, self.frame.shape[0])

    def _dropna(self) -> None:
        """Drop any row that contains a missing value from the frame.

        This is only called if ``drop_na=True`` was supplied to
        :func:`@dispatch() <pdcast.dispatch>`.  Otherwise, missing values will
        be grouped and processed according to a ``CompositeDispatch`` strategy.
        """
        assert self.frame is not None

        original_length = self.frame.shape[0]
        self.frame = self.frame.dropna(how="any")
        self.hasnans = self.frame.shape[0] != original_length

    def _extract_vectors(self, df: pd.DataFrame) -> dict[str, pd.Series]:
        """Split the frame back into individual vectors."""
        # replace the original series name if one was given
        return {
            arg: col.rename(  # type: ignore
                self.series_names.get(arg, None),  # type: ignore
                copy=False
            )
            for arg, col in df.items()
        }

    def _standardize_dtype(
        self,
        vectors: dict[str, pd.Series],
        detected: dict[str, types.VectorType]
    ) -> dict[str, pd.Series]:
        """Standardize each vector's dtype according to the detected type."""
        return {
            arg: series.astype(detected[arg].dtype, copy=False)
            for arg, series in vectors.items()
        }


class DispatchStrategy:
    """An abstract class describing an encapsulated strategy for executing a
    dispatched function.

    These are examples of the Gang of Four `Strategy pattern
    <https://en.wikipedia.org/wiki/Strategy_pattern>`_.

    Parameters
    ----------
    func : DispatchFunc
        A reference to the dispatched function being executed.
    arguments : DispatchArguments
        The arguments to the dispatched function as a
        :class:`DispatchArguments <pdcast.DispatchArguments>` object.
    fill_na : bool
        Indicates whether to replace missing values during the dispatch
        process.

    Notes
    -----
    :class:`DispatchStrategies <pdcast.DispatchStrategy>` are created during
    :meth:`DispatchFunc.__call__() <pdcast.DispatchFunc.__call__>` to handle
    the execution of a dispatched function.  The final strategy is chosen based
    on a combination of external flags managed by the
    :class:`DispatchFunc <pdcast.DispatchFunc>` itself as well as the detected
    types of the dispatched arguments.
    """

    def __init__(
        self,
        func: DispatchFunc,
        arguments: DispatchArguments,
        fill_na: bool
    ):
        self.func = func
        self.signature = func._signature
        self.arguments = arguments
        self.fill_na = fill_na

    def execute(self) -> Any:
        """Execute the dispatched strategy and return the raw result.

        Returns
        -------
        Any
            The result of the dispatched computation without any
            post-processing.  This is always equal to the raw output of the
            dispatched implementation(s).

        Raises
        ------
        NotImplementedError
            If the strategy does not implement an ``execute()`` method.

        Notes
        -----
        :meth:`DispatchStrategy.execute() <pdcast.DispatchStrategy.execute>` and
        :meth:`DispatchStrategy.finalize() <pdcast.DispatchStrategy.finalize>`
        are split into two separate methods to allow strategies to be nested
        within one another.
        
        For instance, a :class:`CompositeDispatch <pdcast.CompositeDispatch>`
        strategy essentially consists of a sequence of
        :class:`HomogenousDispatch <pdcast.HomogenousDispatch>` operations (one
        for each group), where the output is not finalized until the very end.
        Separating the execution and finalization steps allows this to be done
        in a straightforward and efficient manner.
        """
        raise NotImplementedError(
            f"strategy does not implement an `execute()` method: "
            f"{type(self).__qualname__}"
        )

    def finalize(self, result: Any) -> Any:
        """Post-process the raw result of a dispatched computation.

        Parameters
        ----------
        result : Any
            The raw result of the dispatched computation, as returned by
            :meth:`DispatchStrategy.execute() <pdcast.DispatchStrategy.execute>`.

        Returns
        -------
        Any
            The final result of the dispatched computation after applying the
            appropriate post-processing steps.

        Raises
        ------
        NotImplementedError
            If the strategy does not implement a ``finalize()`` method.

        Notes
        -----
        :meth:`DispatchStrategy.execute() <pdcast.DispatchStrategy.execute>` and
        :meth:`DispatchStrategy.finalize() <pdcast.DispatchStrategy.finalize>`
        are split into two separate methods to allow strategies to be nested
        within one another.
        
        For instance, a :class:`CompositeDispatch <pdcast.CompositeDispatch>`
        strategy essentially consists of a sequence of
        :class:`HomogenousDispatch <pdcast.HomogenousDispatch>` operations (one
        for each group), where the output is not finalized until the very end.
        Separating the execution and finalization steps allows this to be done
        in a straightforward and efficient manner.
        """
        raise NotImplementedError(
            f"strategy does not implement a `finalize()` method: "
            f"{type(self).__qualname__}"
        )

    def rectify_series(self, series: pd.Series) -> pd.Series:
        """Normalize the ``dtype`` of a series to its detected type.

        Parameters
        ----------
        series : pd.Series
            The series to rectify.

        Returns
        -------
        pd.Series
            The rectified series.

        Notes
        -----
        This method is called on both the input to and output from a dispatched
        implementation.  Re-labeling the data in this way bypasses a whole
        category of potential bugs that can arise when working with ambiguous
        data, such as those contained within a ``dtype: object`` array or other
        iterable.

        Rectifying the ``dtype`` of an array is also a significant
        optimization, since it allows :func:`detect_type <pdcast.detect_type>`
        to run in constant time throughout the dispatch process, including
        within the dispatched implementations themselves.
        """
        return series.astype(detect_type(series).dtype, copy=False)

    def replace_na(self, series: pd.Series) -> pd.Series:
        """Fill in missing values based on the final index of a dispatched
        series.

        Parameters
        ----------
        series : pd.Series
            The series to pad with missing values.

        Returns
        -------
        pd.Series
            The series with missing values filled in.

        Notes
        -----
        This method is called whenever vectorized input is given to a
        :class:`DispatchFunc <pdcast.DispatchFunc>` with ``fill_na=True``.

        The way it works is by comparing the dispatched implementation's output
        index to that of the original arguments.  Since the input index is
        always normalized to a unique :class:`RangeIndex <pandas.RangeIndex>`,
        it should never contain any gaps between indices.  However, if missing
        values were removed during the dispatch process (via ``drop_na=True``),
        or if the dispatched implementation itself removes values from its
        input, then gaps can be introduced to the index wherever this occurs.
        When this method is called, it searches for these gaps and dynamically
        fills them with the appropriate
        :attr:`na_value <pdcast.ScalarType.na_value>`.

        This allows dispatched implementations to be written without having to
        consider missing values at any point, which can significantly simplify
        their logic in many cases.  What's more, they can inject missing values
        into the final output simply by dropping values from their input, which
        maintains this condition throughout the dispatch process.
        """
        assert self.arguments.original_index is not None

        # get nullable type for series
        nullable = detect_type(series).make_nullable()

        # get length of vectors before normalizing
        original_length = self.arguments.original_index.shape[0]

        # build empty series containing only missing values
        result = pd.Series(
            np.full(original_length, nullable.na_value, dtype=object),
            index = pd.RangeIndex(0, original_length),
            name=series.name,
            # dtype=object  # TODO: decide which of these are best
            dtype=nullable.dtype
        )

        # TODO: When inserting into an ObjectArray, the values will be
        # converted to the output using cast().  By using dtype: object, we
        # can avoid this implicit conversion.  However, this also means that
        # in the case of integers, for instance, we might incur an additional
        # loop to convert the values to astype() them to the appropriate dtype.
        # We should check the performance in both cases for int64 vs int[python]

        # merge result into the empty series
        result[series.index] = series
        # return result.astype(nullable.dtype, copy=False)  # TODO: same as above
        return result


class HomogenousDispatch(DispatchStrategy):
    """Dispatch homogenous inputs to the appropriate implementation.

    This strategy is used to perform a 1-to-1 dispatch between a single group
    of homogenous inputs and a single dispatched implementation.
    """

    def execute(self) -> Any:
        """Search for a dispatched implementation that matches the types of the
        arguments and execute it.

        Returns
        -------
        Any
            The raw result of the dispatched computation, directly from the
            chosen implementation.

        Notes
        -----
        See the :ref:`activity diagram <dispatch.activity_diagram>` for a
        step-by-step overview of the dispatch process.
        """
        bound = self.arguments
        return self.func[bound.key](*bound.args, **bound.kwargs)

    def finalize(self, result: Any) -> Any:
        """Process the result of a dispatched computation, converting it into
        its final output.

        Parameters
        ----------
        result : Any
            The raw result of the dispatched computation, directly from the
            chosen implementation.

        Returns
        -------
        Any
            The final result of the dispatched computation, after applying any
            post-processing steps.

        Notes
        -----
        This method is only called once at the very end of the dispatch
        process.  It is responsible for analyzing the final output and applying
        the correct filter/transform/aggregate logic based on its return type.

        See the :ref:`activity diagram <dispatch.activity_diagram>` for a
        step-by-step overview of the dispatch process.
        """
        # transform
        if isinstance(result, (pd.Series, pd.DataFrame)):
            assert self.arguments.frame is not None
            assert self.arguments.original_index is not None

            # rectify output dtype
            result = self._standardize_dtype(result)

            # check if index is subset of normalized
            self._check_index(result)

            # replace missing values
            frame_length = self.arguments.frame.shape[0]
            if (
                self.fill_na and
                self.arguments.hasnans or result.shape[0] < frame_length
            ):
                result = self._merge_na(result)

            # replace original index
            final_index = self.arguments.original_index
            if not self.fill_na:
                final_index = final_index[result.index]
            result.index = final_index

        # aggregate
        return result

    def _standardize_dtype(
        self,
        result: pd.Series | pd.DataFrame
    ) -> pd.Series | pd.DataFrame:
        """Convert the output ``dtype`` to match its inferred type."""
        # series
        if isinstance(result, pd.Series):
            return self.rectify_series(result)

        # dataframe
        rectified = {
            col: self.rectify_series(series) for col, series in result.items()
        }
        return pd.DataFrame(rectified, copy=False)

    def _check_index(self, result: pd.Series | pd.DataFrame) -> None:
        """Check that the output index is a subset of the original."""
        assert self.arguments.frame is not None

        if not result.index.difference(self.arguments.frame.index).empty:
            sig = [f"{k}: {str(v)}" for k, v in self.arguments.types.items()]
            warn_msg = (
                f"index mismatch in "
                f"'{type(self).__qualname__}({', '.join(sig)})': final index "
                f"is not a subset of starting index"
            )
            warnings.warn(warn_msg, UserWarning, stacklevel=4)

            # index mismatch results in extraneous NaNs
            self.arguments.hasnans = True

    def _merge_na(
        self,
        result: pd.Series | pd.DataFrame
    ) -> pd.Series | pd.DataFrame:
        """Replace any missing indices with the appropriate
        :attr:`na_value <pdcast.ScalarType.na_value>`.
        """
        # series
        if isinstance(result, pd.Series):
            return self.replace_na(result)

        # dataframe
        return pd.DataFrame(
            {col: self.replace_na(series) for col, series in result.items()},
            copy=False
        )


class CompositeDispatch(DispatchStrategy):
    """Dispatch composite inputs to the appropriate implementations.

    This strategy is used to perform many-to-many dispatch between multiple
    groups of composite inputs and a single dispatched implementation for each
    group.
    """

    def __init__(
        self,
        func: DispatchFunc,
        arguments: DispatchArguments,
        fill_na: bool,
        rectify: bool
    ):
        super().__init__(
            func=func,
            arguments=arguments,
            fill_na=fill_na
        )
        self.rectify = rectify

    def execute(self) -> list[tuple[tuple[types.VectorType, ...], Any]]:
        """For each group in the input, search for an appropriate dispatched
        implementation and execute it.

        Returns
        -------
        list[tuple[tuple[types.VectorType, ...], Any]]
            A list of results from the dispatched computations, organized into
            a sequence of ``(key, result)`` tuples.

        Notes
        -----
        See the :ref:`activity diagram <dispatch.activity_diagram>` for a
        step-by-step overview of the dispatch process.
        """
        results = []

        # process each group independently
        for group in self.arguments.groups:
            strategy = HomogenousDispatch(
                self.func,
                arguments=group,
                fill_na=False
            )
            results.append((group.key, strategy.execute()))

        return results

    def finalize(
        self,
        result: list[tuple[tuple[types.VectorType, ...], Any]]
    ) -> Any:
        """Process the result of the dispatched computations, converting them
        into the final output.

        Parameters
        ----------
        result : list[tuple[tuple[types.VectorType, ...], Any]]
            A list containing the raw results of the dispatched computations
            along with their corresponding input keys, packaged as
            ``(key, result)`` tuples.

        Returns
        -------
        Any
            The final result of the dispatched computation, after applying any
            post-processing steps.

        Notes
        -----
        This method is only called once at the very end of the dispatch
        process.  It is responsible for analyzing the final output and applying
        the correct filter/transform/aggregate logic based on its return type.

        See the :ref:`activity diagram <dispatch.activity_diagram>` for a
        step-by-step overview of the dispatch process.
        """
        groups = dict(result)

        # transform
        as_series = all(isinstance(grp, pd.Series) for grp in groups.values())
        as_df = all(isinstance(grp, pd.DataFrame) for grp in groups.values())
        if as_series or as_df:
            # rectify output dtype
            groups = self._standardize_dtype(groups, as_series=as_series)

            # check if index is subset of normalized
            result_index = self._check_index(groups)

            # merge series and replace missing values
            if as_series:
                return self._merge_series(
                    groups,  # type: ignore
                    result_index=result_index
                )

            # returning as dataframe - check for column mismatch
            columns = {tuple(df.columns) for df in groups.values()}
            if len(columns) > 1:
                raise ValueError(
                    f"DataFrames do not share the same column names {columns}"
                )

            # merge each column and replace missing values
            df = {}
            for col in columns.pop():
                df[col] = self._merge_series(
                    {key: df[col] for key, df in groups.items()},
                    result_index=result_index
                )
            return pd.DataFrame(df, copy=False)

        # aggregate
        df = {}
        for idx, arg_name in enumerate(self.arguments.dispatched):
            df[arg_name] = [key[idx] for key in groups]  # type: ignore
        df[f"{self.func.__name__}()"] = list(groups.values())  # type: ignore
        return pd.DataFrame(df, copy=False)

    def _standardize_dtype(
        self,
        groups: dict[tuple[types.VectorType, ...], pd.Series | pd.DataFrame],
        as_series: bool
    ) -> dict[tuple[types.VectorType, ...], pd.Series | pd.DataFrame]:
        """Convert each group's output ``dtype`` to match its inferred type."""
        # series
        if as_series:
            return {
                key: self.rectify_series(grp)  # type: ignore
                for key, grp in groups.items()
            }

        # dataframe
        rectified = {}
        for key, grp in groups.items():
            grp_result = {
                col: self.rectify_series(series) for col, series in grp.items()
            }
            rectified[key] = pd.DataFrame(grp_result, copy=False)
        return rectified  # type: ignore

    def _check_index(
        self,
        groups: dict[tuple[types.VectorType, ...], pd.Series | pd.DataFrame]
    ) -> pd.Index:
        """Validate and merge a collection of transformed Series/DataFrame
        indices, and check that they are a subset of the input index.
        """
        assert self.arguments.frame is not None

        # merge indices
        group_iter = iter(groups.values())
        index = next(group_iter).index
        size = index.shape[0]
        for grp in group_iter:
            index = index.union(grp.index)
            if len(index) < size + grp.shape[0]:
                sig = [
                    f"{k}: {str(v)}" for k, v in self.arguments.types.items()
                ]
                warn_msg = (
                    f"index collision in '{type(self).__qualname__}("
                    f"{', '.join(sig)})': 2 or more implementations returned "
                    f"overlapping indices"
                )
                warnings.warn(warn_msg, UserWarning)
            size = index.shape[0]

        # check that final index is subset of starting index
        if not index.difference(self.arguments.frame.index).empty:
            sig = [f"{k}: {str(v)}" for k, v in self.arguments.types.items()]
            warn_msg = (
                f"index mismatch in '"
                f"{type(self).__qualname__}({', '.join(sig)})': final index "
                f"is not a subset of starting index"
            )
            warnings.warn(warn_msg, UserWarning, stacklevel=4)

            # index mismatch results in extraneous NaNs
            self.arguments.hasnans = True

        # return final index
        return index

    def _merge_series(
        self,
        groups: dict[tuple[types.VectorType, ...], pd.Series],
        result_index: pd.Index
    ) -> pd.Series:
        """Merge the computed series results by index."""
        assert self.arguments.frame is not None
        assert self.arguments.original_index is not None

        # get unique output types
        unique = set(detect_type(series) for series in groups.values())

        # if all types are in same family, attempt to standardize to widest
        if (
            self.rectify and len(unique) > 1 and any(
                all(t2.contains(t1) for t1 in unique)
                for t2 in {t.generic.root for t in unique}
            )
        ):
            from pdcast.convert import cast
            widest = max(unique)  # using comparison operators
            try:
                groups = {
                    key: cast(grp, widest, downcast=False, errors="raise")
                    for key, grp in groups.items()
                }
                unique = {widest}
            except Exception:
                pass

        # results are homogenous
        if len(unique) == 1:
            # merge series
            result = pd.concat(groups.values())  # NOTE: removes ObjectDtypes
            result = result.astype(unique.pop().dtype, copy=False)  # ^^^^^^^
            result.sort_index(inplace=True)

            # replace missing values
            frame_length = self.arguments.frame.shape[0]
            if (
                self.fill_na and
                self.arguments.hasnans or result.shape[0] < frame_length
            ):
                result = self.replace_na(result)

            # replace original index
            final_index = self.arguments.original_index
            if not self.fill_na:
                final_index = final_index[result.index]
            result.index = final_index
            return result

        # results are mixed
        # NOTE: we can't use pd.concat() because it tends to coerce mixed-type
        # results in uncontrollable ways.  Instead, we join each group into a
        # `dtype: object` series to preserve the actual values.

        # generate object series
        if self.fill_na:
            original_length = self.arguments.original_index.shape[0]
            result = pd.Series(
                np.full(original_length, pd.NA, dtype=object),
                index=pd.RangeIndex(0, original_length)
            )
        else:
            result = pd.Series(
                np.full(result_index.shape[0], pd.NA, dtype=object),
                index=result_index  # NOTE: generated in _check_index()
            )

        # join results
        for series in groups.values():
            result[series.index] = series

        # replace with original index
        final_index = self.arguments.original_index
        if not self.fill_na:
            final_index = final_index[result.index]
        result.index = final_index
        return result


def supercedes(
    node1: tuple[types.VectorType, ...],
    node2: tuple[types.VectorType, ...]
) -> bool:
    """Check if ``node1`` is consistent with and strictly more specific than
    ``node2``.

    Parameters
    ----------
    node1 : tuple[types.VectorType, ...]
        The first node to compare.
    node2 : tuple[types.VectorType, ...]
        The second node to compare.

    Returns
    -------
    bool
        ``True`` if ``node1`` is more specific than ``node2``.  ``False``
        otherwise.

    Notes
    -----
    This uses the same hierarchical membership checks as the stand-alone
    :func:`typecheck() <pdcast.typecheck>` function.
    """
    return all(x.contains(y) for x, y in zip(node2, node1))


def edge(
    node1: tuple[types.VectorType, ...],
    node2: tuple[types.VectorType, ...]
) -> bool:
    """Determine if ``node1`` should be checked before ``node2``.

    Parameters
    ----------
    node1 : tuple[types.VectorType, ...]
        The first node to compare.
    node2 : tuple[types.VectorType, ...]
        The second node to compare.

    Returns
    -------
    bool
        ``True`` if ``node1`` should be topologically ordered before ``node2``.
        ``False`` otherwise.

    Notes
    -----
    Ties are broken by recursively backing off the last element of both nodes.
    Whichever one is more specific in its earlier elements (from left to right)
    will always be preferred.
    """
    # pylint: disable=arguments-out-of-order
    if not node1 or not node2:
        return False

    return (
        supercedes(node1, node2) and
        not supercedes(node2, node1) or
        edge(node1[:-1], node2[:-1])  # back off rightmost element
    )


def topological_sort(edges: dict[Any, Any]) -> list[Any]:
    """Topological sort algorithm by Kahn (1962).

    Parameters
    ----------
    edges : dict[Any, Any]
        A dictionary of the form ``{A: {B, C}}`` where ``B`` and ``C`` depend
        on ``A``.

    Returns
    -------
    list[Any]
        An ordered list of nodes that satisfy the dependencies of ``edges``.

    Examples
    --------
    .. doctest::

        >>> topological_sort({1: (2, 3), 2: (3, )})
        [1, 2, 3]

    References
    ----------    
    Kahn, Arthur B. (1962), "Topological sorting of large networks",
    Communications of the ACM
    """
    # edge_count is used to detect cycles
    edge_count = sum(len(dependencies) for dependencies in edges.values())

    # invert edges: {A: {B, C}} -> {B: {A}, C: {A}}
    inverted: dict[Any, Any] = {}
    for node, dependencies in edges.items():
        for dependent in dependencies:
            inverted.setdefault(dependent, set()).add(node)

    # Proceed with Kahn topological sort algorithm
    no_incoming = [node for node in edges if node not in inverted]
    result = []
    while no_incoming:

        # pop a node with no incoming edges (order doesn't matter)
        node = no_incoming.pop()
        result.append(node)

        # for each edge from node -> dependent:
        for dependent in edges.get(node, ()):

            # remove edge from inverted map and decrement edge count
            inverted[dependent].remove(node)
            edge_count -= 1

            # if dependent has no more incoming edges, add it to the queue
            if not inverted[dependent]:
                no_incoming.append(dependent)
                del inverted[dependent]  # no reason to keep an empty set

    # if there are any edges left, then there must be a cycle
    if edge_count:
        cycles = [node for node in edges if inverted.get(node, None)]
        raise ValueError(f"edges are cyclic: {cycles}")

    return result


#######################
####    TESTING    ####
#######################


@dispatch("x", "y", drop_na=False, fill_na=True, rectify=True)
def add(x, y):
    return x + y


@add.overload("int", "int")
def add1(x, y):
    return x - y


@add.overload("null", "int")
def add_null(x, y):
    return pd.Series([256] * len(x), index=x.index)


# @add.overload("int64", "int64")
# def add2(x, y):
#     return x + y


# @add.overload("float", "float")
# def add3(x, y):
#     return x + y



# sig = add._signature
# bound = sig([1, 2, 3.0], [3, 2, 1])

# strat = DirectDispatch(add)




# @dispatch("bar")
# def foo(bar):
#     print("base")
#     return bar

# @foo.overload("int")  # least specific
# def integer_foo(bar):
#     print("integer")
#     return bar

# @foo.overload("int64[numpy]")  # most specific
# def numpy_int64_foo(bar):
#     print("int64[numpy]")
#     return bar

# @foo.overload("int64")
# def int64_foo(bar):
#     print("int64")
#     return bar



"""This module contains various utility functions for interacting with the
proleptic Gregorian calendar.

Functions
---------
date_to_days()
    Convert proleptic Gregorian calendar dates into day offsets from the utc
    epoch.

days_in_month()
    Get the length (in days) of vector of months, accounting for leap years.

days_to_date()
    Convert a vector of day offsets from the utc epoch into proleptic Gregorian
    calendar dates.

is_leap_year()
    Check if a vector of years are leap years according to the proleptic
    Gregorian calendar.

leaps_between()
    Return the number of leap days between 2 years.
"""
import cython
import numpy as np
cimport numpy as np

from pdcast.util.type_hints import array_like, numeric


#########################
####    CONSTANTS    ####
#########################


# Gregorian cycle lengths
cdef unsigned int days_per_400_years = 146097
cdef unsigned short days_per_100_years = 36524
cdef unsigned short days_per_4_years = 1461
cdef unsigned short days_per_year = 365


# raw days per month, starting from January 1st
cdef np.ndarray[np.uint8_t] days_per_month = np.array(
    [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31],
    dtype=np.uint8
)


######################
####    PUBLIC    ####
######################


cpdef object date_to_days(object year, object month, object day):
    """Convert proleptic Gregorian calendar dates into day offsets from the utc
    epoch ('1970-01-01 00:00:00+0000').

    This function works equally well with both scalar and vectorized data,
    which is why it accepts ``object`` types over more specific C integers.

    Parameters
    ----------
    year : int | array-like
        Proleptic Gregorian calendar year.

        .. note:: This function assumes the existence of a year 0, which does
            not correspond to real-world historical dates.  In order to convert
            a historical BC year (`-1 BC`, `-2 BC`, ...) to a negative year
            (`0`, `-1`, ...), simply add one to the BC year.  AD years are
            unaffected.

    month : int | array-like
        Proleptic Gregorian calendar month, indexed from 1 (January).  If a
        month value exceeds the range [1, 12], then any excess is automatically
        carried over into `year`.
    day : int | array-like
        Proleptic Gregorian calendar day, indexed from 1.  If a day value
        exceeds the maximum for the selected month, any excess is automatically
        carried over into `month` and `year`.

    Returns
    -------
    int | array-like:
        An integer day offset from the utc epoch.

    See Also
    --------
    days_to_date : UTC day offset to Gregorian calendar date.

    Notes
    -----
    Each of this function's arguments can be vectorized, and any excess beyond
    the normal range ([1-12] for `month`, [1-31] for `day`) is accurately
    reflected in the returned day offset.  For example, to measure the length
    (in days) of a 1 month period starting on December 31st, 2000, one can
    simply call:

        >>> date_to_days(2000, 12 + 1, 31) - date_to_days(2000, 12, 31)
        31

    .. note:: for the sake of efficiency, this function will not attempt to
        coerce numpy integers or integer arrays into their built-in python
        equivalents.  As such, they may silently overflow (and wrap around
        infinity) if memory limits are exceeded during conversion.  This
        shouldn't be a problem in practice: even with day-level precision, the
        valid 64-bit range vastly exceeds the observed age of the universe.
        Nevertheless, this can be avoided by converting the inputs into python
        integers (which do not overflow) beforehand.

    References
    ----------
    http://www.algonomicon.org/calendar/gregorian/to_jdn.html
    https://stackoverflow.com/questions/11609769/efficient-algorithm-for-converting-number-of-days-to-years-including-leap-years

    Examples
    --------
    Inputs can be realistic:

    >>> date_to_days(1970, 1, 1)
    0
    >>> date_to_days(1970, 1, 2)
    1
    >>> date_to_days(1970, 2, 1)
    31
    >>> date_to_days(1970, 3, 1)
    59
    >>> date_to_days(1968, 1, 1)
    -731
    >>> date_to_days(1969, 1, 1)
    -365
    >>> date_to_days(1971, 1, 1)
    365
    >>> date_to_days(1972, 1, 1)
    730

    Or unrealistic:

    >>> date_to_days(1970, 1, 0)
    -1
    >>> date_to_days(1970, 1, -1)
    -2
    >>> date_to_days(1970, 0, 1)
    -31
    >>> date_to_days(1970, -1, 1)
    -61

    And potentially vectorized:

    >>> date_to_days(1970, 1, np.arange(-1, 3))
    array([-2, -1,  0,  1])
    >>> date_to_days(1970, np.arange(-1, 3), 1)
    array([-61, -31,   0,  31])
    >>> date_to_days(np.arange(1968, 1973), 1, 1)
    array([-731, -365,    0,  365,  730])
    """
    cdef object result

    # normalize months to start with March 1st, indexed from 1 (January)
    month = month - 3
    year = year + month // 12
    day = day - 1
    month %= 12

    # build result
    result = 365 * year + year // 4 - year // 100 + year // 400
    result += (153 * month + 2) // 5
    result += day
    result -= 719468  # move origin from March 1st, year 0 to utc
    return result


cpdef object days_in_month(object month, object year = 2001):
    """Get the length in days of a given month, taking leap years into
    account.

    This function works equally well with both scalar and vectorized data,
    which is why it accepts ``object`` types over more specific C integers.

    .. warning:: This should not be used with indices outside the range 1
        (January) to 12 (December).

    Parameters
    ----------
    month : int | array-like
        The month in question, indexed from 1 (January).  Can be vectorized.
    year : int | array-like
        The year to consider for each month.  Relevant for determining the
        length of February, which varies between 28 and 29 based on this
        setting.

    Returns
    -------
    int | array-like
        The number of days in the given month or a vector of month lengths.

    Raises
    ------
    ValueError
        If `month` isn't between -11 and 12.  Negative values have to be
        accepted due to python's native support for negative indexing, which
        cannot easily be disabled.  Nevertheless, only values between 1
        (January) and 12 (December) should be used in practice.

    Examples
    --------
    Inputs can be scalar:

    >>> days_in_month(1, 2001)
    31
    >>> days_in_month(2, 2001)
    28
    >>> days_in_month(2, 2000)
    29

    Or vectorized:

    >>> days_in_month(np.arange(-2, 3), 2001)
    array([31, 30, 31, 31, 28], dtype=uint8)
    """
    if hasattr(month, "astype"):
        month = month.astype(np.int8, copy=False)
    try:
        return days_per_month[month - 1] + ((month == 2) & is_leap_year(year))
    except IndexError as err:
        err_msg = (
            f"`month` must be between 1 (January) and 12 (December), not "
            f"{month}"
        )
        raise ValueError(err_msg) from err


cpdef dict days_to_date(object days):
    """Convert day offsets from the utc epoch ('1970-01-01 00:00:00+0000') into
    proleptic Gregorian calendar dates.

    This function works equally well with both scalar and vectorized data,
    which is why it accepts ``object`` types over more specific C integers.

    Parameters
    ----------
    days : int | array-like
        An integer day offset from the utc epoch or a vector of such offsets.  

    Returns
    -------
    dict[str, int | array-like]
        A dictionary with the following keys/values:
            * `'year'`: Proleptic Gregorian calendar year.
            * `'month'`: Proleptic Gregorian calendar month, indexed from 1.
            * `'day'`: Proleptic Gregorian calendar day, indexed from 1.

        .. note:: This function assumes the existence of a year 0, which does
            not correspond to real-world historical dates.  In order to convert
            a historical BC year (`-1 BC`, `-2 BC`, ...) to a negative year
            (`0`, `-1`, ...), simply add one to the BC year.  AD years are
            unaffected.

    See Also
    --------
    date_to_days : Gregorian calendar date to UTC day offset.

    Notes
    -----
    This is the inverse operation of
    :func:`date_to_days <pdcast.util.time.date_to_days>`.  Its output can be
    used as a **kwargs dict for that function.

    .. note:: for the sake of efficiency, this function will not attempt to
        coerce numpy integers or integer arrays into their built-in python
        equivalents.  As such, they may silently overflow (and wrap around
        infinity) if 64-bit limits are exceeded during conversion.  This
        shouldn't be a problem in practice; even with day-level precision, the
        valid 64-bit range vastly exceeds the observed age of the universe.
        Nevertheless, this can be avoided by converting the inputs into python
        integers (which do not overflow) beforehand.

    References
    ----------
    http://www.algonomicon.org/calendar/gregorian/from_jdn.html
    https://stackoverflow.com/questions/11609769/efficient-algorithm-for-converting-number-of-days-to-years-including-leap-years

    Examples
    --------
    Day offsets can be positive or negative:

    >>> days_to_date(0)
    {'year': 1970, 'month': 1, 'day': 1}
    >>> days_to_date(1)
    {'year': 1970, 'month': 1, 'day': 2}
    >>> days_to_date(-1)
    {'year': 1969, 'month': 12, 'day': 31}

    With correct month lengths:

    >>> days_to_date(31)
    {'year': 1970, 'month': 2, 'day': 1}
    >>> days_to_date(31 + 28)
    {'year': 1970, 'month': 3, 'day': 1}
    >>> days_to_date(365 + 365 + 31 + 28)  # leap day
    {'year': 1972, 'month': 2, 'day': 29}

    And correct year lengths:

    >>> days_to_date(365)
    {'year': 1971, 'month': 1, 'day': 1}
    >>> days_to_date(365 + 365)
    {'year': 1972, 'month': 1, 'day': 1}
    >>> days_to_date(365 + 365 + 365)  # leap year
    {'year': 1972, 'month': 12, 'day': 31}

    They can also be vectorized:

    >>> days_to_date(np.arange(-1, 2))
    {'year': array([1969, 1970, 1970]), 'month': array([12,  1,  1]), 'day': array([31,  1,  2])}
    """
    cdef object years
    cdef object months

    # move origin from utc to March 1st, 2000 (start of 400-year cycle)
    days = days - 11017

    # count years since 2000, correcting for leaps in each Gregorian cycle
    years = days - (days + 1) // days_per_400_years  # 400-year cycles
    years += years // days_per_100_years  # 100-year cycles
    years -= (years + 1) // days_per_4_years  # 4-year cycles
    years //= days_per_year  # whole years

    # compute residual days in final year
    days -= days_per_year * years + years // 4 - years // 100 + years // 400

    # exploit symmetry in month lengths, starting from March 1st
    months = (5 * days + 2) // 153  # no need for a lookup table
    days -= (153 * months + 2) // 5  # residual days in month

    # convert to proper calendar year
    years += 2000 + (months >= 10)  # treat Jan, Feb as belonging to next year

    # undo bias toward march 1st
    months += 2
    months %= 12

    # index from 1
    months += 1
    days += 1

    # return as dict
    return {"year": years, "month": months, "day": days}

cpdef inline object is_leap_year(object year):
    """Check if the given year is a leap year according to the proleptic
    Gregorian calendar.

    This function works equally well with both scalar and vectorized data,
    which is why it accepts ``object`` types over more specific C integers.

    Parameters
    ----------
    year : int | array-like
        The year to be checked.

    Returns
    -------
    bool | array-like
        `True` where `year` is a leap year.  `False` otherwise.

    Examples
    --------
    Years can be scalar:

    >>> is_leap_year(1971)
    False
    >>> is_leap_year(1972)
    True

    Or vectorized:

    >>> is_leap_year(np.arange(1968, 1973))
    array([ True, False, False, False,  True])

    This function can also be used to obtain the length of a particular year.

    >>> 365 + is_leap_year(1972)
    366
    """
    return (year % 4 == 0) & ((year % 100 != 0) | (year % 400 == 0))


cpdef inline object leaps_between(object lower, object upper):
    """Return the number of leap days between the years `lower` and `upper`.

    This function works equally well with both scalar and vectorized data,
    which is why it accepts ``object`` types over more specific C integers.

    Parameters
    ----------
    lower : int | array-like
        The starting year to consider or a vector of starting years.
    upper : int | array-like
        The ending year to consider or a vector of ending years.

    Returns
    -------
    int
        The number of leap days between each index of `lower` and `upper`.

    Notes
    -----
    This function counts from the beginning of each year.  This means that
    `leaps_between(x, x + 1)` will return 1 if and only if `x` was a leap year.

    Identical to `calendar.leapdays()` from the built-in `calendar` package,
    but avoids an import and is compiled (and thus slightly faster).

    Examples
    --------
    Bounds can be scalar:

    >>> leaps_between(1968, 1970)  # 1968 was a leap year
    1
    >>> leaps_between(1968, 1974)  # 1972 was also a leap year
    2
    >>> leaps_between(0, 2022)
    491

    Or vectorized:

    >>> leaps_between(1970 + np.arange(-6, 6), 2000)
    array([9, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6])
    """
    # start from beginning of year
    upper = upper - 1
    lower = lower - 1

    cdef object up = upper // 4 - upper // 100 + upper // 400
    cdef object down = lower // 4 - lower // 100 + lower // 400

    return up - down



"""This package defines the structure and contents of the bertrand type system,
including the resolve() and detect() helper functions as well as abstract base
classes that can extend it to arbitrary data.
"""
from .base import *
from .boolean import Bool, NumpyBool, PandasBool, PythonBool
from .categorical import Categorical
from .complex import (
    Complex, Complex64, NumpyComplex64, Complex128, NumpyComplex128, PythonComplex,
    Complex160, NumpyComplex160
)
# from .datetime import (
#     Datetime, PandasTimestamp, PythonDatetime, NumpyDatetime64
# )
from .decimal import Decimal, PythonDecimal
from .float import (
    Float, Float16, NumpyFloat16, Float32, NumpyFloat32, Float64, NumpyFloat64,
    PythonFloat, Float80, NumpyFloat80
)
from .integer import (
    Int, Signed, PythonInt, Int64, NumpyInt64, PandasInt64, Int32, NumpyInt32,
    PandasInt32, Int16, NumpyInt16, PandasInt16, Int8, NumpyInt8, PandasInt8, Unsigned,
    UInt64, NumpyUInt64, PandasUInt64, UInt32, NumpyUInt32, PandasUInt32, UInt16,
    NumpyUInt16, PandasUInt16, UInt8, NumpyUInt8, PandasUInt8
)
from .missing import Missing
from .object import Object
from .sparse import Sparse
from .string import String, PythonString, PyArrowString
from .timedelta import Timedelta, PandasTimedelta, PythonTimedelta, NumpyTimedelta64


# importing * from base also masks module names, which can be troublesome
# pylint: disable=undefined-variable
del meta  # type: ignore



"""This module contains helper functions for conversions to and from string
representations.
"""


# TODO: deprecate this function in favor of full regex matching
cpdef char boolean_match(
    str val,
    dict lookup,
    bint ignore_case,
    char fill
) except -1:
    """Convert a scalar string into a boolean value using the given lookup
    dictionary.

    Notes
    -----
    The ``fill`` argument is implemented as an 8-bit integer as an
    optimization.  It can take on the following values:

        *   ``1``, signaling that ``KeyErrors`` should be taken as ``True``
        *   ``0``, signaling that ``KeyErrors`` should be taken as ``False``
        *   ``-1``, signaling that ``KeyErrors`` should be raised up the stack.

    """
    if ignore_case:
        val = val.lower()

    if fill == -1:
        try:
            return lookup[val]
        except KeyError as err:
            err_msg = f"encountered non-boolean value: '{val}'"
            raise ValueError(err_msg) from err

    return lookup.get(val, fill)


cdef Py_UNICODE[36] base_lookup = (
    [chr(ord("0") + i) for i in range(10)] + 
    [chr(ord("A") + i) for i in range(26)]
)


cpdef str int_to_base(object val, unsigned char base):
    """Convert an integer into a string with the given base."""
    cdef bint negative
    cdef list chars
    cdef str result

    if not val:
        return "0"

    negative = val < 0
    if negative:
        val = abs(val)

    chars = []
    while val:
        chars.append(base_lookup[val % base])
        val //= base

    result = "".join(chars[::-1])
    if negative:
        result = "-" + result
    return result



"""This module contains all the prepackaged complex types for the ``pdcast``
type system.
"""
import numpy as np

from .base import Type, TypeMeta


class Complex(Type):
    """Abstract complex type."""

    aliases = {"complex", "cfloat", "complex float", "complex floating", "c"}

    @property
    def equiv_float(self) -> TypeMeta:
        """An equivalent floating point type."""
        from .float import Float
        return Float


@Complex.default
class Complex128(Complex):
    """Abstract 128-bit complex type."""

    aliases = {"complex128", "cdouble", "complex double", "complex_", "c16", "D"}
    max = 2**53
    min = -2**53

    @property
    def equiv_float(self) -> TypeMeta:
        """An equivalent floating point type."""
        from .float import Float64
        return Float64


@Complex128.default
class NumpyComplex128(Complex128, backend="numpy"):
    """Numpy 128-bit complex type."""

    aliases = {np.complex128, np.dtype(np.complex128)}
    dtype = np.dtype(np.complex128)
    missing = np.complex128("nan+nanj")

    @property
    def equiv_float(self) -> TypeMeta:
        """An equivalent floating point type."""
        from .float import NumpyFloat64
        return NumpyFloat64


class PythonComplex(Complex128, backend="python"):
    """Python complex type."""

    aliases = {complex}
    scalar = complex
    missing = np.complex128("nan+nanj")

    @property
    def equiv_float(self) -> TypeMeta:
        """An equivalent floating point type."""
        from .float import PythonFloat
        return PythonFloat


class Complex64(Complex):
    """Abstract 64-bit complex type."""

    aliases = {"complex64", "csingle", "complex single", "singlecomplex", "c8", "F"}
    max = 2**24
    min = -2**24

    @property
    def equiv_float(self) -> TypeMeta:
        """An equivalent floating point type."""
        from .float import Float32
        return Float32


@Complex64.default
class NumpyComplex64(Complex64, backend="numpy"):
    """Numpy 64-bit complex type."""

    aliases = {np.complex64, np.dtype(np.complex64)}
    dtype = np.dtype(np.complex64)
    missing = np.complex64("nan+nanj")

    @property
    def equiv_float(self) -> TypeMeta:
        """An equivalent floating point type."""
        from .float import NumpyFloat32
        return NumpyFloat32


#########################################
####    COMPLEX LONG DOUBLE (x86)    ####
#########################################


# TODO: figure out conditional types


# NOTE: long doubles are platform-specific and may not be valid depending on
# hardware configuration.


has_clongdouble: bool = np.dtype(np.clongdouble).itemsize > 16


class Complex160(Complex):
    """Abstract 160-bit float type."""

    aliases = {
        "complex160", "clongdouble", "clongfloat", "complex longdouble",
        "complex longfloat", "complex long double", "complex long float",
        "longcomplex", "long complex", "c20", "G"
    }
    max = 2**64
    min = -2**64

    @property
    def equiv_float(self) -> TypeMeta:
        """An equivalent floating point type."""
        from .float import Float80
        return Float80


@Complex160.default
class NumpyComplex160(Complex160, backend="numpy"):
    """Numpy 160-bit complex type."""

    aliases = {np.clongdouble, np.dtype(np.clongdouble)}
    dtype = np.dtype(np.clongdouble)
    missing = np.clongdouble("nan+nanj")

    @property
    def equiv_float(self) -> TypeMeta:
        """An equivalent floating point type."""
        from .float import NumpyFloat80
        return NumpyFloat80



"""This package holds the glue necessary to attach and detach
``pdcast``-related functionality to/from pandas data structures, as well as
to dispatch method calls based on type.

Functions
---------
attach()
    globally attach all ``pdcast`` functionality to ``pandas.Series`` and
    ``pandas.DataFrame`` objects.

detach()
    globally detach all ``pdcast`` functionality from ``pandas.Series`` and
    ``pandas.DataFrame`` objects, returning them to their original state.

Classes
-------
Namespace
    A virtual namespace for ``pandas.Series`` instances.  These are created
    on the fly by the ``@dispatch`` decorator.

DispatchMethod
    A decorator for a series method that automatically dispatches to an
    implementation of the appropriate type.
"""
